"""Tests for converting WiX Localization (.wxl) files to/from Gettext PO."""

from io import BytesIO

from lxml import etree

from translate.convert import po2wxl, wxl2po
from translate.storage import wxl

from . import test_convert

WXL_V4_TEMPLATE = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de" Codepage="65001">
  <String Id="WixUIBack" Overridable="yes" Value="Back" />
  <UI Id="WixUI_Mondo" />
  <String Id="WixUICancel" Overridable="yes" Value="Cancel" />
  <String Id="WixUINext" Overridable="yes" Value="Next" />
</WixLocalization>
"""

WXL_V3_TEMPLATE = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://schemas.microsoft.com/wix/2006/localization" Culture="cs-cz">
  <String Id="WixUIBack">Back</String>
  <String Id="WixUICancel">Cancel</String>
</WixLocalization>
"""

PO_CONTENT = """
#: WixUIBack
msgid "WixUIBack"
msgstr "Zurueck"

#: WixUICancel
msgid "WixUICancel"
msgstr "Abbrechen"
"""


class TestWxl2PO:
    WXL_V4 = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de" Codepage="1252">
  <String Id="WixUIBack" Overridable="yes" Value="Zurueck" />
  <String Id="WixUICancel" Overridable="yes" Value="Abbrechen" />
</WixLocalization>
"""

    WXL_V3 = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://schemas.microsoft.com/wix/2006/localization" Culture="pt-br" Codepage="1252">
  <String Id="WixUIBack">Voltar</String>
  <String Id="WixUICancel">Cancelar</String>
</WixLocalization>
"""

    WXL_EN = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="en-us">
  <String Id="WixUIBack" Value="Back" />
  <String Id="WixUICancel" Value="Cancel" />
  <String Id="WixUINext" Value="Next" />
</WixLocalization>
"""

    @staticmethod
    def _convert(input_bytes):
        inputfile = BytesIO(input_bytes)
        outputfile = BytesIO()
        converter = wxl2po.wxl2po(inputfile, outputfile)
        result = converter.run()
        return result, converter.target_store, outputfile

    @staticmethod
    def _convert_with_template(input_bytes, template_bytes):
        inputfile = BytesIO(input_bytes)
        outputfile = BytesIO()
        templatefile = BytesIO(template_bytes)
        converter = wxl2po.wxl2po(inputfile, outputfile, templatefile)
        result = converter.run()
        return result, converter.target_store, outputfile

    def test_convert_v4_value_attribute(self) -> None:
        """WiX v4 Value-attribute strings are converted to PO units."""
        result, store, _ = self._convert(self.WXL_V4)
        assert result == 1
        # skip header
        units = [u for u in store.units if not u.isheader()]
        assert len(units) == 2
        assert store.findunit("WixUIBack") is not None
        assert store.findunit("WixUIBack").target == "Zurueck"

    def test_convert_v3_text_content(self) -> None:
        """WiX v3 text-content strings are converted to PO units."""
        result, store, _ = self._convert(self.WXL_V3)
        assert result == 1
        units = [u for u in store.units if not u.isheader()]
        assert len(units) == 2
        assert store.findunit("WixUIBack").target == "Voltar"
        assert store.findunit("WixUICancel").target == "Cancelar"

    def test_empty_file_returns_zero(self) -> None:
        """An empty WXL file yields a return code of 0."""
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de" />
"""
        result, _, _ = self._convert(content)
        assert result == 0

    def test_output_contains_po_header(self) -> None:
        """Converted PO output contains the standard PO header."""
        _, _, outputfile = self._convert(self.WXL_V4)
        output = outputfile.getvalue().decode("utf-8")
        assert "Project-Id-Version" in output
        assert 'msgid "WixUIBack"' in output

    def test_merge_with_template(self) -> None:
        """With a template, msgid comes from template and msgstr from input."""
        wxl_de = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de">
  <String Id="WixUIBack" Value="Zurueck" />
  <String Id="WixUICancel" Value="Abbrechen" />
</WixLocalization>
"""
        result, store, _ = self._convert_with_template(wxl_de, self.WXL_EN)
        assert result == 1
        units = [u for u in store.units if not u.isheader()]
        # All template keys appear in output
        assert len(units) == 3
        back_unit = store.findunit("WixUIBack")
        assert back_unit is not None
        assert back_unit.target == "Zurueck"
        cancel_unit = store.findunit("WixUICancel")
        assert cancel_unit is not None
        assert cancel_unit.target == "Abbrechen"

    def test_merge_template_missing_translation(self) -> None:
        """Keys in template but absent in input appear with empty msgstr."""
        wxl_de = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de">
  <String Id="WixUIBack" Value="Zurueck" />
</WixLocalization>
"""
        result, store, _ = self._convert_with_template(wxl_de, self.WXL_EN)
        assert result == 1
        units = [u for u in store.units if not u.isheader()]
        # All three template keys appear
        assert len(units) == 3
        # WixUINext has no translation in input → empty msgstr
        next_unit = store.findunit("WixUINext")
        assert next_unit is not None
        assert next_unit.target == ""

    def test_merge_non_translatable_ui_excluded(self) -> None:
        """Non-translatable UI elements (no Text attr) are excluded from PO output."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="en-us">
  <String Id="WixUIBack" Value="Back" />
  <UI Id="WixUI_Mondo" />
  <String Id="WixUINext" Value="Next" />
</WixLocalization>
"""
        wxl_de = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de">
  <String Id="WixUIBack" Value="Zurueck" />
  <String Id="WixUINext" Value="Weiter" />
</WixLocalization>
"""
        result, store, outputfile = self._convert_with_template(wxl_de, template)
        assert result == 1
        units = [u for u in store.units if not u.isheader()]
        # WixUI_Mondo (non-translatable) must not appear as a PO unit
        assert len(units) == 2
        po_text = outputfile.getvalue().decode("utf-8")
        assert "WixUI_Mondo" not in po_text
        assert 'msgid "WixUIBack"' in po_text
        assert 'msgstr "Zurueck"' in po_text

    def test_merge_xml_comments_excluded(self) -> None:
        """XML comments in the WXL template are excluded from PO output."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="en-us">
  <!-- Buttons -->
  <String Id="WixUIBack" Value="Back" />
  <!-- Navigation -->
  <String Id="WixUINext" Value="Next" />
</WixLocalization>
"""
        wxl_de = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de">
  <String Id="WixUIBack" Value="Zurueck" />
  <String Id="WixUINext" Value="Weiter" />
</WixLocalization>
"""
        result, _store, outputfile = self._convert_with_template(wxl_de, template)
        assert result == 1
        po_text = outputfile.getvalue().decode("utf-8")
        assert "Buttons" not in po_text
        assert "Navigation" not in po_text
        assert 'msgid "WixUIBack"' in po_text
        assert 'msgstr "Zurueck"' in po_text
        assert 'msgid "WixUINext"' in po_text
        assert 'msgstr "Weiter"' in po_text

    def test_merge_translatable_ui_with_text_included(self) -> None:
        """UI elements with a Text attribute are treated as translatable."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="en-us">
  <UI Id="WixUIInstallTitle" Text="Installation" />
  <UI Id="WixUI_Mondo" />
</WixLocalization>
"""
        wxl_de = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de">
  <UI Id="WixUIInstallTitle" Text="Installation DE" />
</WixLocalization>
"""
        result, store, outputfile = self._convert_with_template(wxl_de, template)
        assert result == 1
        units = [u for u in store.units if not u.isheader()]
        # WixUI_Mondo has no Text → not translatable, so only 1 unit
        assert len(units) == 1
        po_text = outputfile.getvalue().decode("utf-8")
        assert 'msgid "WixUIInstallTitle"' in po_text
        assert 'msgstr "Installation DE"' in po_text
        assert "WixUI_Mondo" not in po_text

    def test_no_template_non_translatable_ui_excluded(self) -> None:
        """Without a template, non-translatable UI elements are excluded from PO."""
        content = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de">
  <!-- A comment -->
  <String Id="WixUIBack" Value="Zurueck" />
  <UI Id="WixUI_Mondo" />
  <UI Id="WixUIInstallTitle" Text="Installation DE" />
</WixLocalization>
"""
        result, store, outputfile = self._convert(content)
        assert result == 1
        units = [u for u in store.units if not u.isheader()]
        assert len(units) == 2
        po_text = outputfile.getvalue().decode("utf-8")
        assert "WixUI_Mondo" not in po_text
        assert "A comment" not in po_text
        assert 'msgid "WixUIBack"' in po_text
        assert 'msgid "WixUIInstallTitle"' in po_text

    def test_merge_v3_template(self) -> None:
        """Template handling works with WiX v3 text-content format."""
        wxl_en_v3 = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://schemas.microsoft.com/wix/2006/localization" Culture="en-us">
  <String Id="WixUIBack">Back</String>
  <String Id="WixUICancel">Cancel</String>
</WixLocalization>
"""
        result, store, _ = self._convert_with_template(self.WXL_V3, wxl_en_v3)
        assert result == 1
        back_unit = store.findunit("WixUIBack")
        assert back_unit is not None
        assert back_unit.target == "Voltar"
        cancel_unit = store.findunit("WixUICancel")
        assert cancel_unit is not None
        assert cancel_unit.target == "Cancelar"


class TestPO2WXL:
    PO_SOURCE = b"""
msgid "WixUIBack"
msgstr "Voltar"

msgid "WixUICancel"
msgstr "Cancelar"
"""

    @staticmethod
    def _convert(po_bytes, template_bytes=None):
        inputfile = BytesIO(po_bytes)
        outputfile = BytesIO()
        templatefile = BytesIO(template_bytes) if template_bytes else None
        converter = po2wxl.po2wxl(inputfile, outputfile, templatefile)
        result = converter.run()
        return result, converter.target_store, outputfile

    def _convert_to_string(self, po_bytes, template_bytes=None):
        _, _, outputfile = self._convert(po_bytes, template_bytes)
        return outputfile.getvalue().decode("utf-8")

    def test_basic_conversion(self) -> None:
        """PO units are written as WXL String elements."""
        output = self._convert_to_string(self.PO_SOURCE)
        assert "WixUIBack" in output
        assert "Voltar" in output
        assert "WixUICancel" in output
        assert "Cancelar" in output

    def test_output_is_valid_xml(self) -> None:
        """po2wxl output is valid XML."""
        output_bytes = BytesIO(self.PO_SOURCE)
        out = BytesIO()
        po2wxl.po2wxl(output_bytes, out).run()
        # Should not raise
        etree.fromstring(out.getvalue())

    def test_with_template(self) -> None:
        """When a template is provided, extra attributes are preserved."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de" Codepage="1252">
  <String Id="WixUIBack" Overridable="yes" Value="" />
  <String Id="WixUICancel" Overridable="yes" Value="" />
</WixLocalization>
"""
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"

msgid "WixUICancel"
msgstr "Abbrechen"
"""
        output = self._convert_to_string(po, template)
        assert 'Overridable="yes"' in output
        assert "Zurueck" in output
        assert "Abbrechen" in output

    def test_culture_from_template(self) -> None:
        """Culture attribute is preserved from the template."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="cs-cz">
  <String Id="Key" Value="" />
</WixLocalization>
"""
        po = b"""
msgid "Key"
msgstr "Hodnota"
"""
        output = self._convert_to_string(po, template)
        assert 'Culture="cs-cz"' in output

    def test_template_preserves_xml_comments(self) -> None:
        """XML comments in the template are preserved in po2wxl WXL output."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="en-us">
  <!-- Navigation buttons -->
  <String Id="WixUIBack" Value="Back" />
  <!-- Footer -->
  <String Id="WixUINext" Value="Next" />
</WixLocalization>
"""
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"

msgid "WixUINext"
msgstr "Weiter"
"""
        output = self._convert_to_string(po, template)
        assert "<!-- Navigation buttons -->" in output
        assert "<!-- Footer -->" in output
        assert "Zurueck" in output
        assert "Weiter" in output

    def test_template_preserves_non_translatable_ui(self) -> None:
        """Non-translatable UI elements (no Text attr) are preserved in po2wxl WXL output."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="en-us">
  <String Id="WixUIBack" Value="Back" />
  <UI Id="WixUI_Mondo" />
  <String Id="WixUINext" Value="Next" />
</WixLocalization>
"""
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"

msgid "WixUINext"
msgstr "Weiter"
"""
        output = self._convert_to_string(po, template)
        assert 'Id="WixUI_Mondo"' in output
        assert "Zurueck" in output
        assert "Weiter" in output

    def test_template_output_matches_structure(self) -> None:
        """po2wxl output with template is structurally identical to template with changed values."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="en-us">
  <!-- Buttons -->
  <String Id="WixUIBack" Value="Back" />
  <UI Id="WixUI_Mondo" />
</WixLocalization>
"""
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"
"""
        output = self._convert_to_string(po, template)
        # Comment preserved
        assert "<!-- Buttons -->" in output
        # Non-translatable UI preserved
        assert 'Id="WixUI_Mondo"' in output
        # Translation applied
        assert 'Value="Zurueck"' in output
        # Source value replaced (not present)
        assert 'Value="Back"' not in output

    def test_empty_po_returns_zero(self) -> None:
        """An empty PO file yields a return code of 0."""
        result, _, _ = self._convert(b"")
        assert result == 0

    def test_comments_preserved(self) -> None:
        """XML comments in the template are preserved in the output."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de" Codepage="65001">
  <!-- Back button label -->
  <String Id="WixUIBack" Value="Back" />
  <!-- Cancel button label -->
  <String Id="WixUICancel" Value="Cancel" />
</WixLocalization>
"""
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"

msgid "WixUICancel"
msgstr "Abbrechen"
"""
        output = self._convert_to_string(po, template)
        assert "<!-- Back button label -->" in output
        assert "<!-- Cancel button label -->" in output
        assert "Zurueck" in output
        assert "Abbrechen" in output

    def test_nontranslatable_ui_preserved(self) -> None:
        """Non-translatable UI elements in the template are preserved."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de" Codepage="65001">
  <String Id="WixUIBack" Value="Back" />
  <UI Id="WixUI_Advanced" />
  <UI Id="WixUI_Minimal" ControlType="Button" />
</WixLocalization>
"""
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"
"""
        output = self._convert_to_string(po, template)
        assert 'Id="WixUI_Advanced"' in output
        assert 'Id="WixUI_Minimal"' in output
        assert 'ControlType="Button"' in output

    def test_untranslated_keeps_template_value(self) -> None:
        """Strings without a translation keep the value from the template."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de" Codepage="65001">
  <String Id="WixUIBack" Value="Back" />
  <String Id="WixUICancel" Value="Cancel" />
</WixLocalization>
"""
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"
"""
        _, store, _ = self._convert(po, template)
        assert store.findid("WixUIBack").target == "Zurueck"
        assert store.findid("WixUICancel").target == "Cancel"

    def test_untranslated_no_template(self) -> None:
        """Without a template, untranslated units produce an empty string."""
        po = b"""
msgid "WixUIBack"
msgstr ""
"""
        _, store, _ = self._convert(po)
        unit = store.findid("WixUIBack")
        assert unit is not None
        assert unit.target == ""

    def test_element_order_preserved(self) -> None:
        """Element ordering from the template is preserved in the output."""
        template = b"""<?xml version="1.0" encoding="utf-8"?>
<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl" Culture="de-de" Codepage="65001">
  <String Id="ZZZ" Value="Last" />
  <String Id="AAA" Value="First" />
</WixLocalization>
"""
        po = b"""
msgid "ZZZ"
msgstr "ZZZ-translated"

msgid "AAA"
msgstr "AAA-translated"
"""
        output = self._convert_to_string(po, template)
        assert output.index("ZZZ") < output.index("AAA")


class TestPO2WXLCommand(test_convert.TestConvertCommand):
    """Tests running actual po2wxl commands on files."""

    convertmodule = po2wxl
    defaultoptions = {"progress": "none"}
    expected_options = [
        "-t TEMPLATE, --template=TEMPLATE",
    ]

    def test_convert(self) -> None:
        """Tests the conversion to a WXL file using a template."""
        self.create_testfile("template.wxl", WXL_V4_TEMPLATE)
        self.create_testfile("input.po", PO_CONTENT)
        self.run_command(template="template.wxl", i="input.po", o="output.wxl")
        with self.open_testfile("output.wxl") as handle:
            content = handle.read()
        wxl_result = wxl.WxlFile(BytesIO(content))
        assert wxl_result.findid("WixUIBack").target == "Zurueck"
        assert wxl_result.findid("WixUICancel").target == "Abbrechen"
        assert 'Id="WixUI_Mondo"' in content.decode("utf-8")

    def test_preserve_comments(self) -> None:
        """XML comments in the template are preserved in the output."""
        template_with_comments = (
            b'<?xml version="1.0" encoding="utf-8"?>\n'
            b'<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl"'
            b' Culture="de-de" Codepage="65001">\n'
            b"  <!-- Back button label -->\n"
            b'  <String Id="WixUIBack" Value="Back" />\n'
            b"  <!-- Cancel button label -->\n"
            b'  <String Id="WixUICancel" Value="Cancel" />\n'
            b"</WixLocalization>\n"
        )
        self.create_testfile("template.wxl", template_with_comments)
        self.create_testfile("input.po", PO_CONTENT)
        self.run_command(template="template.wxl", i="input.po", o="output.wxl")
        with self.open_testfile("output.wxl") as handle:
            content = handle.read().decode("utf-8")
        assert "<!-- Back button label -->" in content
        assert "<!-- Cancel button label -->" in content
        assert "Zurueck" in content
        assert "Abbrechen" in content

    def test_preserve_nontranslatable_ui(self) -> None:
        """Non-translatable UI elements in the template are preserved."""
        template_with_ui = (
            b'<?xml version="1.0" encoding="utf-8"?>\n'
            b'<WixLocalization xmlns="http://wixtoolset.org/schemas/v4/wxl"'
            b' Culture="de-de" Codepage="65001">\n'
            b'  <String Id="WixUIBack" Value="Back" />\n'
            b'  <UI Id="WixUI_Advanced" />\n'
            b'  <UI Id="WixUI_Minimal" ControlType="Button" />\n'
            b"</WixLocalization>\n"
        )
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"
"""
        self.create_testfile("template.wxl", template_with_ui)
        self.create_testfile("input.po", po)
        self.run_command(template="template.wxl", i="input.po", o="output.wxl")
        with self.open_testfile("output.wxl") as handle:
            content = handle.read().decode("utf-8")
        assert 'Id="WixUI_Advanced"' in content
        assert 'Id="WixUI_Minimal"' in content
        assert 'ControlType="Button"' in content

    def test_untranslated_kept_from_template(self) -> None:
        """Untranslated strings use the value from the template."""
        self.create_testfile("template.wxl", WXL_V4_TEMPLATE)
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"
"""
        self.create_testfile("input.po", po)
        self.run_command(template="template.wxl", i="input.po", o="output.wxl")
        with self.open_testfile("output.wxl") as handle:
            wxl_result = wxl.WxlFile(handle)
        assert wxl_result.findid("WixUIBack").target == "Zurueck"
        assert wxl_result.findid("WixUINext").target == "Next"

    def test_convert_v3(self) -> None:
        """V3 text-content format is handled correctly with a template."""
        self.create_testfile("template.wxl", WXL_V3_TEMPLATE)
        po = b"""
msgid "WixUIBack"
msgstr "Zurueck"

msgid "WixUICancel"
msgstr "Abbrechen"
"""
        self.create_testfile("input.po", po)
        self.run_command(template="template.wxl", i="input.po", o="output.wxl")
        with self.open_testfile("output.wxl") as handle:
            wxl_result = wxl.WxlFile(handle)
        assert wxl_result.findid("WixUIBack").target == "Zurueck"
        assert wxl_result.findid("WixUICancel").target == "Abbrechen"



#
# Copyright 2004-2006,2008 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#

"""module for parsing html files for translation."""

import html.parser
import re
from html.entities import html5

from translate.lang.data import is_rtl
from translate.storage import base
from translate.storage.base import ParseError

# Override the piclose tag from simple > to ?> otherwise we consume HTML
# within the processing instructions
html.parser.piclose = re.compile(r"\?>")  # ty:ignore[unresolved-attribute]


class htmlunit(base.TranslationUnit):
    """A unit of translatable/localisable HTML content."""

    def __init__(self, source=None) -> None:
        super().__init__(source)
        self.locations = []
        self._context = ""

    def addlocation(self, location) -> None:
        self.locations.append(location)

    def getlocations(self):
        """Get the list of locations for this unit."""
        return self.locations

    def getcontext(self):
        """Get the message context."""
        return self._context

    def setcontext(self, context) -> None:
        """Set the message context."""
        self._context = context or ""

    def getid(self):
        """Returns a unique identifier for this unit."""
        if self._context:
            return f"{self._context}\04{self.source}"
        return self.source


class htmlfile(html.parser.HTMLParser, base.TranslationStore):
    UnitClass = htmlunit

    TRANSLATABLE_ELEMENTS = [
        "address",
        "article",
        "aside",
        "blockquote",
        "button",
        "caption",
        "dd",
        "dt",
        "div",
        "figcaption",
        "footer",
        "header",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "label",
        "li",
        "main",
        "nav",
        "option",
        "p",
        "pre",
        "section",
        "td",
        "th",
        "title",
    ]
    """These HTML elements (tags) will be extracted as translation units, unless
    they lack translatable text content.
    In case one translatable element is embedded in another, the outer translation
    unit will be split into the parts before and after the inner translation unit."""

    TRANSLATABLE_ATTRIBUTES = [
        "abbr",  # abbreviation for a table header cell
        "alt",
        "lang",  # only for the html element -- see extract_translatable_attributes()
        "summary",
        "title",  # tooltip text for an element
        "value",
    ]
    """Text from these HTML attributes will be extracted as translation units.
    Note: the content attribute of meta tags is a special case."""

    TRANSLATABLE_METADATA = [
        "description",
        "keywords",
        "og:title",
        "og:description",
        "og:site_name",
        "og:image:alt",
        "twitter:title",
        "twitter:description",
        "twitter:image:alt",
        "video:actor:role",
        "video:tag",
        "article:section",
        "article:tag",
        "payment:description",
    ]
    """Document metadata from meta elements with these names will be extracted as translation units.
    Includes standard meta tags and common social media tags (Open Graph and Twitter Cards).
    Reference `<https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta/name>`_"""

    EMPTY_HTML_ELEMENTS = [
        "area",
        "base",
        "br",
        "col",
        "embed",
        "hr",
        "img",
        "input",
        "link",
        "meta",
        "param",
        "source",
        "track",
        "wbr",
    ]
    """An empty element is an element that cannot have any child nodes (i.e., nested
    elements or text nodes). In HTML, using a closing tag on an empty element is
    usually invalid.
    Reference `<https://developer.mozilla.org/en-US/docs/Glossary/Empty_element>`_"""

    WHITESPACE_RE = re.compile(r"\s+")

    LEADING_WHITESPACE_RE = re.compile(r"^(\s+)")

    TRAILING_WHITESPACE_RE = re.compile(r"(\s+)$")

    ENCODING_RE = re.compile(
        rb"""<meta.*
                                content.*=.*?charset=\s*?
                                ([^\s]*)
                                \s*?["']\s*?>
                             """,
        re.VERBOSE | re.IGNORECASE,
    )

    def __init__(self, inputfile=None, callback=None) -> None:
        super().__init__(convert_charrefs=False)
        base.TranslationStore.__init__(self)

        # store parameters
        self.filename = getattr(inputfile, "name", None)
        if callback is None:
            self.callback = self._simple_callback
        else:
            self.callback = callback

        # initialize state
        self.filesrc = ""
        self.tag_path = []
        self.tu_content = []
        self.tu_location = None
        self.tu_docpath = None
        self.ignore_depth = 0  # Track nesting level of ignored elements
        self.comment_ignore = False  # Track if inside <!-- translate:off -->
        self.ignore_tag_stack = []  # Track which tags have ignore attribute
        self.ancestor_id_stack = []  # stack of ancestor ids (top is nearest)
        self._id_pushed_stack = []  # parallel to tag stack: did this element push an id?
        self._id_depth_stack = []  # indices into tag_path where ids were pushed
        self._id_pos_stack = []  # positions (line, col) where ancestor ids start
        self._ancestor_id_label_stack = []  # labels for ancestor ids (id or id:line-col)
        self._id_seen = set()  # track seen ids to disambiguate labels
        self._units_by_source = {}  # Track units by normalized source to add context only when needed for disambiguation
        self._units_by_src_ctx = {}  # Fast lookup for units created with explicit context: (source, context) -> unit
        # Stack of dicts for sibling tag counts at each nesting level (for docpath)
        self._sibling_counts = [{}]
        # Stack of (tag, index) tuples for current docpath
        self._docpath_stack = []
        # Track translated language value when <html lang=""> is translated, used to sync og:locale
        self._translated_lang = None

        # parse
        if inputfile is not None:
            htmlsrc = inputfile.read()
            inputfile.close()
            self.parse(htmlsrc)

    @staticmethod
    def _simple_callback(string):
        return string

    def guess_encoding(self, htmlsrc):
        """
        Returns the encoding of the html text.

        We look for 'charset=' within a meta tag to do this.
        """
        result = self.ENCODING_RE.findall(htmlsrc)
        if result:
            self.encoding = result[0].decode("ascii")
        return self.encoding

    def do_encoding(self, htmlsrc):
        """Return the html text properly encoded based on a charset."""
        self.guess_encoding(htmlsrc)
        return htmlsrc.decode(self.encoding)

    def parse(self, htmlsrc) -> None:  # ty:ignore[invalid-method-override]
        # Reset translation state for new parse
        self._translated_lang = None
        htmlsrc = self.do_encoding(htmlsrc)
        self.feed(htmlsrc)

    def is_extraction_ignored(self):
        """Check if we're currently in an ignored section."""
        return self.ignore_depth > 0 or self.comment_ignore

    def _build_docpath(self) -> str:
        """Build the current document path from the docpath stack."""
        return "/".join(f"{tag}[{index}]" for tag, index in self._docpath_stack)

    def begin_translation_unit(self) -> None:
        # at the start of a translation unit:
        # this interrupts any translation unit in progress, so process the queue
        # and prepare for the new.
        self.emit_translation_unit()
        self.tu_content = []
        self.tu_location = f"{self.filename}+{'.'.join(self.tag_path)}:{self.getpos()[0]}-{self.getpos()[1] + 1}"
        self.tu_docpath = self._build_docpath()

    def end_translation_unit(self) -> None:
        # at the end of a translation unit:
        # process the queue and reset state.
        self.emit_translation_unit()
        self.tu_content = []
        self.tu_location = None
        self.tu_docpath = None

    def append_markup(self, markup) -> None:
        # if within a translation unit: add to the queue to be processed later.
        # otherwise handle immediately.
        # However, if we're in an ignored section, output directly to filesrc
        # to avoid including ignored content in parent translation units
        if self.is_extraction_ignored():
            self.filesrc += markup["html_content"]
        elif self.tu_location:
            self.tu_content.append(markup)
        else:
            self.emit_attribute_translation_units(markup)
            self.filesrc += markup["html_content"]

    def emit_translation_unit(self) -> None:
        # If we're in an ignored section, just output the raw content
        if self.is_extraction_ignored():
            for markup in self.tu_content:
                self.filesrc += markup["html_content"]
            return

        # scan through the queue:
        # - find the first and last translatable markup elements: the captured
        #   interval [start, end)
        # - match start and end tags
        start = 0
        end = 0
        tagstack = []
        tagmap = {}
        tag = None
        do_normalize = True
        for pos, content in enumerate(self.tu_content):
            if content["type"] != "endtag" and tag in self.EMPTY_HTML_ELEMENTS:
                match = tagstack.pop()
                tag = None

            if self.has_translatable_content(content):
                if end == 0:
                    start = pos
                end = pos + 1
            elif content["type"] == "starttag":
                tagstack.append(pos)
                tag = content["tag"]
                if tag == "pre":
                    do_normalize = False
            elif content["type"] == "endtag":
                if tagstack:
                    match = tagstack.pop()
                    tagmap[match] = pos
                    tagmap[pos] = match
                tag = None

        # if no translatable content found: process all the content in the queue
        # as if the translation unit didn't exist.
        if end == 0:
            for markup in self.tu_content:
                self.emit_attribute_translation_units(markup)
                self.filesrc += markup["html_content"]
            return

        # scan the start and end tags captured between translatable content;
        # extend the captured interval to include the matching tags
        for pos in range(start + 1, end - 1):
            if self.tu_content[pos]["type"] in {"starttag", "endtag"} and pos in tagmap:
                match = tagmap[pos]
                start = min(start, match)
                end = max(end, match + 1)

        # emit leading uncaptured markup elements
        for markup in self.tu_content[0:start]:
            if markup["type"] != "comment":
                self.emit_attribute_translation_units(markup)
                self.filesrc += markup["html_content"]

        # emit captured markup elements
        if start < end:
            html_content = []
            for markup in self.tu_content[start:end]:
                if markup["type"] != "comment":
                    if "untranslated_html" in markup:
                        html_content.append(markup["untranslated_html"])
                    else:
                        html_content.append(markup["html_content"])
            html_content = "".join(html_content)
            if do_normalize:
                normalized_content = self.WHITESPACE_RE.sub(" ", html_content.strip())
            else:
                normalized_content = html_content.strip()
            assert normalized_content  # shouldn't be here otherwise

            # Determine context from data-translate-context (outermost wins)
            # tu_content starts at the unit's opening tag, so the first
            # translate_context encountered is the correct one.
            explicit_context = next(
                (
                    m["translate_context"]
                    for m in self.tu_content
                    if "translate_context" in m
                ),
                None,
            )

            # If an explicit context is present, reuse an existing unit with the
            # same source+context instead of creating a duplicate unit.
            unit = None
            if explicit_context:
                unit = self._units_by_src_ctx.get(
                    (normalized_content, explicit_context)
                )
            if unit is None:
                unit = self.addsourceunit(normalized_content)
                if explicit_context:
                    unit.setcontext(explicit_context)
                    # Register for faster lookups
                    self._units_by_src_ctx[normalized_content, explicit_context] = unit

            unit.addlocation(self.tu_location)
            if self.tu_docpath:
                unit.setdocpath(self.tu_docpath)

            # If no explicit context, capture a context hint (from id/ancestor) for potential disambiguation
            context_hint = None
            if not explicit_context:
                context_hint = next(
                    (m["context_hint"] for m in self.tu_content if "context_hint" in m),
                    None,
                )
                # Register the unit for potential disambiguation using context hints
                self._register_unit_for_disambiguation(
                    unit, normalized_content, context_hint
                )

            # Extract comment text from HTML comment elements within the translation unit
            comments = [
                markup["note"]
                for markup in self.tu_content
                if markup["type"] == "comment"
            ]

            # Extract translator comments from data-translate-comment attributes on any tags within the translation unit
            translate_comments = [
                markup["translate_comment"]
                for markup in self.tu_content
                if "translate_comment" in markup
            ]

            # Combine and add all comments to the unit
            all_comments = comments + translate_comments
            if all_comments:
                unit.addnote("\n".join(all_comments), origin="source code")

            html_content = (
                self.get_leading_whitespace(html_content)
                + self.callback(normalized_content)
                + self.get_trailing_whitespace(html_content)
            )
            self.filesrc += html_content

        # emit trailing uncaptured markup elements
        for markup in self.tu_content[end : len(self.tu_content)]:
            if markup["type"] != "comment":
                self.emit_attribute_translation_units(markup)
                self.filesrc += markup["html_content"]

    @staticmethod
    def has_translatable_content(markup):
        # processing instructions count as translatable content, because PHP
        return markup["type"] in {"data", "pi"} and markup["html_content"].strip()

    def extract_translatable_attributes(self, tag, attrs):
        # Don't extract attributes if we're in an ignored section
        if self.is_extraction_ignored():
            return []

        result = []
        if tag == "meta":
            tu = self.create_metadata_attribute_tu(attrs)
            if tu:
                result.append(tu)
        else:
            for attrname, attrvalue in attrs:
                if (
                    attrname in self.TRANSLATABLE_ATTRIBUTES
                    and self.translatable_attribute_matches_tag(attrname, tag)
                ):
                    tu = self.create_attribute_tu(attrname, attrvalue)
                    if tu:
                        result.append(tu)
        return result

    def create_metadata_attribute_tu(self, attrs):
        attrs_dict = dict(attrs)
        # Check both 'name' and 'property' attributes (Open Graph uses 'property')
        name = attrs_dict.get("name", "").lower()
        if not name:
            name = attrs_dict.get("property", "").lower()
        if name in self.TRANSLATABLE_METADATA and "content" in attrs_dict:
            return self.create_attribute_tu("content", attrs_dict["content"])
        return None

    @staticmethod
    def translatable_attribute_matches_tag(attrname, tag):
        if attrname == "lang":
            return tag == "html"
        return True

    def create_attribute_tu(self, attrname, attrvalue):
        normalized_value = self.WHITESPACE_RE.sub(" ", attrvalue).strip()
        if normalized_value:
            return {
                "html_content": normalized_value,
                "attrname": attrname,
                "location": f"{self.filename}+{'.'.join(self.tag_path)}[{attrname}]:{self.getpos()[0]}-{self.getpos()[1] + 1}",
                "docpath": self._build_docpath() + f"[{attrname}]",
            }
        return None

    def emit_attribute_translation_units(self, markup) -> None:
        if "attribute_tus" in markup:
            for tu in markup["attribute_tus"]:
                # Attribute context: prefer explicit translate_context; otherwise use context_hint only if needed
                attr_suffix = f"[{tu.get('attrname')}]" if tu.get("attrname") else ""
                explicit = markup.get("translate_context")
                full_explicit = f"{explicit}{attr_suffix}" if explicit else None

                # If explicit context exists, reuse an existing unit (source+context)
                unit = None
                if full_explicit:
                    unit = self._units_by_src_ctx.get(
                        (tu["html_content"], full_explicit)
                    )
                if unit is None:
                    unit = self.addsourceunit(tu["html_content"])
                    if full_explicit:
                        unit.setcontext(full_explicit)
                        # Register for faster lookups
                        self._units_by_src_ctx[tu["html_content"], full_explicit] = unit
                unit.addlocation(tu["location"])
                if tu.get("docpath"):
                    unit.setdocpath(tu["docpath"])

                if not full_explicit:
                    hint_base = markup.get("context_hint")
                    hint = f"{hint_base}{attr_suffix}" if hint_base else None
                    self._register_unit_for_disambiguation(
                        unit, tu["html_content"], hint
                    )

    def translate_attributes(self, tag, attrs):
        if not attrs:
            return []
        result = []
        translated_lang = None

        # First pass: check if lang is being translated on html tag
        if tag == "html":
            attrs_dict = dict(attrs)
            if "lang" in attrs_dict:
                normalized_value = self.WHITESPACE_RE.sub(
                    " ", attrs_dict["lang"]
                ).strip()
                translated_value = self.callback(normalized_value)
                if translated_value != normalized_value:
                    translated_lang = translated_value
                    # Store translated language for og:locale synchronization
                    self._translated_lang = translated_value

        for attrname, attrvalue in attrs:
            # When translating the lang attribute on the <html> tag, we intentionally discard
            # any existing dir attribute and set it based on the translated language below.
            # This ensures dir reflects the text directionality of the target language.
            if tag == "html" and attrname == "dir" and translated_lang:
                continue
            if attrvalue:
                # Special handling for meta tag content attribute
                if tag == "meta" and attrname == "content":
                    # Check both 'name' and 'property' attributes
                    attrs_dict = dict(attrs)
                    name = attrs_dict.get("name", "").lower()
                    if not name:
                        name = attrs_dict.get("property", "").lower()
                    # Automatically synchronize og:locale with the translated language when
                    # the <html lang=""> attribute was translated during this parse.
                    # This ensures consistency between the page language and Open Graph metadata.
                    # Similar to how dir is automatically set, this intentionally overrides
                    # any explicit translation to maintain consistency.
                    if name == "og:locale" and self._translated_lang:
                        result.append((attrname, self._translated_lang))
                        continue
                    if name in self.TRANSLATABLE_METADATA:
                        normalized_value = self.WHITESPACE_RE.sub(
                            " ", attrvalue
                        ).strip()
                        translated_value = self.callback(normalized_value)
                        if translated_value != normalized_value:
                            result.append((attrname, translated_value))
                            continue
                # Only translate attributes that are translatable for this specific tag
                if (
                    attrname in self.TRANSLATABLE_ATTRIBUTES
                    and self.translatable_attribute_matches_tag(attrname, tag)
                ):
                    normalized_value = self.WHITESPACE_RE.sub(" ", attrvalue).strip()
                    translated_value = self.callback(normalized_value)
                    if translated_value != normalized_value:
                        result.append((attrname, translated_value))
                        continue
            result.append((attrname, attrvalue))

        # Set dir attribute on html tag when lang is being translated
        if tag == "html" and translated_lang:
            result.append(("dir", "rtl" if is_rtl(translated_lang) else "ltr"))

        return result

    @staticmethod
    def create_start_tag(tag, attrs=None, startend=False) -> str:
        attr_strings = []
        attrs = attrs or []
        for attrname, attrvalue in attrs:
            if attrvalue is None:
                attr_strings.append(f" {attrname}")
            else:
                attr_strings.append(f' {attrname}="{attrvalue}"')
        return f"<{tag}{''.join(attr_strings)}{' /' if startend else ''}>"

    def _pop_docpath_level(self) -> None:
        """Pop the current docpath level (sibling counts and path entry)."""
        if self._docpath_stack:
            self._docpath_stack.pop()
        if len(self._sibling_counts) > 1:
            self._sibling_counts.pop()

    def auto_close_empty_element(self) -> None:
        if self.tag_path and self.tag_path[-1] in self.EMPTY_HTML_ELEMENTS:
            self.tag_path.pop()
            self._pop_docpath_level()

    def get_leading_whitespace(self, text: str):
        match = self.LEADING_WHITESPACE_RE.search(text)
        return match.group(1) if match else ""

    def get_trailing_whitespace(self, text: str):
        match = self.TRAILING_WHITESPACE_RE.search(text)
        return match.group(1) if match else ""

    # From here on below, follows the methods of the HTMLParser

    def handle_starttag(self, tag, attrs) -> None:
        self.auto_close_empty_element()
        self.tag_path.append(tag)

        # Update docpath tracking: count siblings and push new level
        counts = self._sibling_counts[-1]
        counts[tag] = counts.get(tag, 0) + 1
        self._docpath_stack.append((tag, counts[tag]))
        self._sibling_counts.append({})

        # Check for data-translate-ignore attribute
        attrs_dict = dict(attrs)
        has_ignore_attr = "data-translate-ignore" in attrs_dict

        if has_ignore_attr:
            self.ignore_depth += 1
            self.ignore_tag_stack.append(tag)

        # Only begin translation unit if not in ignored section
        if tag in self.TRANSLATABLE_ELEMENTS and not self.is_extraction_ignored():
            self.begin_translation_unit()

        translated_attrs = self.translate_attributes(tag, attrs)
        markup = {
            "type": "starttag",
            "tag": tag,
            "html_content": self.create_start_tag(tag, translated_attrs),
            "untranslated_html": self.create_start_tag(tag, attrs),
            "attribute_tus": self.extract_translatable_attributes(tag, attrs),
        }

        # Extract data-translate-comment attribute
        if "data-translate-comment" in attrs_dict:
            markup["translate_comment"] = attrs_dict["data-translate-comment"]

        # Extract data-translate-context attribute
        if "data-translate-context" in attrs_dict:
            context_raw = attrs_dict.get("data-translate-context") or ""
            context_value = self.WHITESPACE_RE.sub(" ", context_raw).strip()
            if context_value:
                markup["translate_context"] = context_value
            # Even with explicit context, record that no id was pushed
            self._id_pushed_stack.append(False)
        elif "id" in attrs_dict:
            # Fallback: use element id as context if present, prefixed with filename
            id_raw = attrs_dict.get("id") or ""
            id_value = self.WHITESPACE_RE.sub(" ", id_raw).strip()
            if id_value:
                id_line, id_col = self.getpos()[0], self.getpos()[1] + 1
                # Determine label for ancestor path hints
                if id_value in getattr(self, "_id_seen", set()):
                    id_label = f"{id_value}:{id_line}-{id_col}"
                    markup["context_hint"] = (
                        f"{self.filename}+{id_value}:{id_line}-{id_col}"
                    )
                else:
                    id_label = id_value
                    markup["context_hint"] = f"{self.filename}:{id_value}"
                    self._id_seen.add(id_value)
                self.ancestor_id_stack.append(id_value)
                self._ancestor_id_label_stack.append(id_label)
                self._id_pushed_stack.append(True)
                self._id_depth_stack.append(len(self.tag_path) - 1)
                self._id_pos_stack.append((id_line, id_col))
        else:
            # No explicit context and no own id; if within ancestor with id, use path
            ancestor_id = self.ancestor_id_stack[-1] if self.ancestor_id_stack else None
            if ancestor_id:
                line, col = self.getpos()[0], self.getpos()[1] + 1
                # Build relative tag path from nearest id ancestor depth
                id_depth = self._id_depth_stack[-1] if self._id_depth_stack else -1
                rel_tags = self.tag_path[id_depth + 1 :]
                ancestor_label = (
                    self._ancestor_id_label_stack[-1]
                    if self._ancestor_id_label_stack
                    else ancestor_id
                )
                path = (
                    ".".join([ancestor_label, *rel_tags])
                    if rel_tags
                    else ancestor_label
                )
                ancestor_line, ancestor_col = (
                    self._id_pos_stack[-1] if self._id_pos_stack else (1, 1)
                )
                rel_line = line - ancestor_line
                rel_col = col - ancestor_col
                markup["context_hint"] = f"{self.filename}+{path}:{rel_line}-{rel_col}"
            self._id_pushed_stack.append(False)

        self.append_markup(markup)

    def handle_endtag(self, tag) -> None:
        try:
            popped = self.tag_path.pop()
        except IndexError:
            raise ParseError(
                f"Mismatched tags: no more tags: line {self.getpos()[0]}"
            ) from None
        if popped != tag and popped in self.EMPTY_HTML_ELEMENTS:
            popped = self.tag_path.pop()
            self._pop_docpath_level()
        if popped != tag:
            raise ParseError(
                "Mismatched closing tag: "
                f"expected '{popped}' got '{tag}' at line {self.getpos()[0]}"
            )

        self._pop_docpath_level()

        self.append_markup({"type": "endtag", "html_content": f"</{tag}>"})

        # Check if this closing tag corresponds to an ignored tag
        # We match from the end of the stack because tags are properly nested
        if self.ignore_tag_stack and self.ignore_tag_stack[-1] == tag:
            self.ignore_tag_stack.pop()
            self.ignore_depth -= 1

        if tag in self.TRANSLATABLE_ELEMENTS and not self.is_extraction_ignored():
            self.end_translation_unit()
            if any(t in self.TRANSLATABLE_ELEMENTS for t in self.tag_path):
                self.begin_translation_unit()

        # Pop ancestor id if closing the element that defined it
        if self._id_pushed_stack:
            pushed = self._id_pushed_stack.pop()
            if pushed and self.ancestor_id_stack:
                self.ancestor_id_stack.pop()
                if self._id_depth_stack:
                    self._id_depth_stack.pop()
                if self._id_pos_stack:
                    self._id_pos_stack.pop()
                if self._ancestor_id_label_stack:
                    self._ancestor_id_label_stack.pop()

    def handle_startendtag(self, tag, attrs) -> None:
        self.auto_close_empty_element()
        self.tag_path.append(tag)

        # Update docpath tracking: count siblings and push new level
        counts = self._sibling_counts[-1]
        counts[tag] = counts.get(tag, 0) + 1
        self._docpath_stack.append((tag, counts[tag]))
        self._sibling_counts.append({})

        # Check for data-translate-ignore attribute
        attrs_dict = dict(attrs)
        has_ignore_attr = "data-translate-ignore" in attrs_dict

        # For self-closing tags with ignore attribute, temporarily set ignore state
        if has_ignore_attr:
            self.ignore_depth += 1

        if tag in self.TRANSLATABLE_ELEMENTS and not self.is_extraction_ignored():
            self.begin_translation_unit()

        translated_attrs = self.translate_attributes(tag, attrs)
        markup = {
            "type": "startendtag",
            "html_content": self.create_start_tag(tag, translated_attrs, startend=True),
            "untranslated_html": self.create_start_tag(tag, attrs, startend=True),
            "attribute_tus": self.extract_translatable_attributes(tag, attrs),
        }

        # Extract data-translate-comment attribute
        if "data-translate-comment" in attrs_dict:
            markup["translate_comment"] = attrs_dict["data-translate-comment"]

        # Extract data-translate-context attribute
        if "data-translate-context" in attrs_dict:
            context_raw = attrs_dict.get("data-translate-context") or ""
            context_value = self.WHITESPACE_RE.sub(" ", context_raw).strip()
            if context_value:
                markup["translate_context"] = context_value
                self._id_pushed_stack.append(False)
        elif "id" in attrs_dict:
            id_raw = attrs_dict.get("id") or ""
            id_value = self.WHITESPACE_RE.sub(" ", id_raw).strip()
            if id_value:
                id_line, id_col = self.getpos()[0], self.getpos()[1] + 1
                if id_value in getattr(self, "_id_seen", set()):
                    id_label = f"{id_value}:{id_line}-{id_col}"
                    markup["context_hint"] = (
                        f"{self.filename}+{id_value}:{id_line}-{id_col}"
                    )
                else:
                    id_label = id_value
                    markup["context_hint"] = f"{self.filename}:{id_value}"
                    self._id_seen.add(id_value)
                # temporary ancestor; will be popped below
                self.ancestor_id_stack.append(id_value)
                self._ancestor_id_label_stack.append(id_label)
                self._id_pushed_stack.append(True)
                self._id_depth_stack.append(len(self.tag_path) - 1)
                self._id_pos_stack.append((id_line, id_col))
        else:
            # No explicit context and no own id; if within ancestor with id, use path as hint
            ancestor_id = self.ancestor_id_stack[-1] if self.ancestor_id_stack else None
            if ancestor_id:
                line, col = self.getpos()[0], self.getpos()[1] + 1
                id_depth = self._id_depth_stack[-1] if self._id_depth_stack else -1
                rel_tags = self.tag_path[id_depth + 1 :]
                ancestor_label = (
                    self._ancestor_id_label_stack[-1]
                    if self._ancestor_id_label_stack
                    else ancestor_id
                )
                path = (
                    ".".join([ancestor_label, *rel_tags])
                    if rel_tags
                    else ancestor_label
                )
                ancestor_line, ancestor_col = (
                    self._id_pos_stack[-1] if self._id_pos_stack else (1, 1)
                )
                rel_line = line - ancestor_line
                rel_col = col - ancestor_col
                markup["context_hint"] = f"{self.filename}+{path}:{rel_line}-{rel_col}"
            self._id_pushed_stack.append(False)

        self.append_markup(markup)

        if tag in self.TRANSLATABLE_ELEMENTS and not self.is_extraction_ignored():
            self.end_translation_unit()
            if any(t in self.TRANSLATABLE_ELEMENTS for t in self.tag_path):
                self.begin_translation_unit()

        # Restore ignore state if we set it
        if has_ignore_attr:
            self.ignore_depth -= 1

        self.tag_path.pop()
        self._pop_docpath_level()
        # For startend, if we pushed an id, pop it now
        if self._id_pushed_stack:
            pushed = self._id_pushed_stack.pop()
            if pushed and self.ancestor_id_stack:
                self.ancestor_id_stack.pop()
                if self._id_depth_stack:
                    self._id_depth_stack.pop()
                if self._id_pos_stack:
                    self._id_pos_stack.pop()
                if self._ancestor_id_label_stack:
                    self._ancestor_id_label_stack.pop()

    def _register_unit_for_disambiguation(self, unit, source, context_hint) -> None:
        bucket = self._units_by_source.setdefault(source, [])
        bucket.append((unit, context_hint))
        if len(bucket) == 2:
            for u, hint in bucket:
                if hint and not u.getcontext():
                    u.setcontext(hint)
        elif len(bucket) > 2 and context_hint and not unit.getcontext():
            unit.setcontext(context_hint)

    def handle_data(self, data) -> None:
        self.auto_close_empty_element()
        self.append_markup({"type": "data", "html_content": data})

    def handle_charref(self, name) -> None:
        """Handle entries in the form &#NNNN; e.g. &#8417;."""
        if name.lower().startswith("x"):
            self.handle_data(chr(int(name[1:], 16)))
        else:
            self.handle_data(chr(int(name)))

    def handle_entityref(self, name) -> None:
        """Handle named entities of the form &aaaa; e.g. &rsquo;."""
        converted = html5.get(f"{name};")
        if name in {"gt", "lt", "amp"} or not converted:
            self.handle_data(f"&{name};")
        else:
            self.handle_data(converted)

    def handle_comment(self, data) -> None:
        self.auto_close_empty_element()

        # Check for translate:off and translate:on directives
        stripped_data = data.strip()
        if stripped_data == "translate:off":
            self.comment_ignore = True
        elif stripped_data == "translate:on":
            self.comment_ignore = False

        self.append_markup(
            {"type": "comment", "html_content": f"<!--{data}-->", "note": data}
        )

    def handle_decl(self, decl) -> None:
        self.auto_close_empty_element()
        self.append_markup({"type": "decl", "html_content": f"<!{decl}>"})

    def handle_pi(self, data) -> None:
        self.auto_close_empty_element()
        self.append_markup({"type": "pi", "html_content": f"<?{data}?>"})

    def unknown_decl(self, data) -> None:
        self.auto_close_empty_element()
        self.append_markup({"type": "cdecl", "html_content": f"<![{data}]>"})


class POHTMLParser(htmlfile):
    pass



from translate.lang import factory


def test_punctranslate() -> None:
    """Tests that we can translate punctuation."""
    language = factory.getlanguage("vi")
    assert language.punctranslate("") == ""
    assert language.punctranslate("abc efg") == "abc efg"
    assert language.punctranslate("abc efg.") == "abc efg."
    assert language.punctranslate("abc efg!") == "abc efg !"
    assert language.punctranslate("abc efg? hij!") == "abc efg? hij !"
    assert language.punctranslate("Delete file: %s?") == "Delete file : %s?"
    assert language.punctranslate('The user "root"') == "The user «\u00a0root\u00a0»"
    # More exhaustive testing of the quoting is in test_fr.py
    assert language.punctranslate('Lưu "Tập tin"') == "Lưu «\u00a0Tập tin\u00a0»"
    assert language.punctranslate("Lưu 'Tập tin'") == "Lưu «\u00a0Tập tin\u00a0»"
    assert language.punctranslate("Lưu `Tập tin'") == "Lưu «\u00a0Tập tin\u00a0»"


def test_sentences() -> None:
    """Tests basic functionality of sentence segmentation."""
    language = factory.getlanguage("vi")
    sentences = language.sentences("")
    assert sentences == []

    sentences = language.sentences("Normal case. Nothing interesting.")
    assert sentences == ["Normal case.", "Nothing interesting."]
    sentences = language.sentences("Is that the case ? Sounds interesting !")
    assert sentences == ["Is that the case ?", "Sounds interesting !"]



from translate.lang import factory


def test_sentences() -> None:
    """Tests basic functionality of sentence segmentation."""
    language = factory.getlanguage("tr")
    sentences = language.sentences("Normal case. Nothing interesting.")
    assert sentences == ["Normal case.", "Nothing interesting."]
    sentences = language.sentences("1. sayı, 2. sayı.")
    assert sentences == ["1. sayı, 2. sayı."]



from translate.storage.placeables import general


def test_placeable_numbers() -> None:
    """Check the correct functioning of number placeables."""
    result = general.NumberPlaceable.parse("Here is a 25 number")
    assert result
    # pylint: disable-next=unsupported-membership-test
    assert general.NumberPlaceable(["25"]) in result
    result = general.NumberPlaceable.parse("Here is a -25 number")
    assert result
    # pylint: disable-next=unsupported-membership-test
    assert general.NumberPlaceable(["-25"]) in result
    result = general.NumberPlaceable.parse("Here is a +25 number")
    assert result
    # pylint: disable-next=unsupported-membership-test
    assert general.NumberPlaceable(["+25"]) in result
    result = general.NumberPlaceable.parse("Here is a 25.00 number")
    assert result
    # pylint: disable-next=unsupported-membership-test
    assert general.NumberPlaceable(["25.00"]) in result
    result = general.NumberPlaceable.parse("Here is a 2,500.00 number")
    assert result
    # pylint: disable-next=unsupported-membership-test
    assert general.NumberPlaceable(["2,500.00"]) in result
    result = general.NumberPlaceable.parse("Here is a 1\u00a0000,99 number")
    assert result
    # pylint: disable-next=unsupported-membership-test
    assert general.NumberPlaceable(["1\u00a0000,99"]) in result


def test_placeable_newline() -> None:
    result = general.NewlinePlaceable.parse("A newline\n")
    assert result
    assert result[1] == general.NewlinePlaceable(["\n"])
    result = general.NewlinePlaceable.parse("First\nSecond")
    assert result
    assert result[1] == general.NewlinePlaceable(["\n"])


def test_placeable_alt_attr() -> None:
    result = general.AltAttrPlaceable.parse(
        'Click on the <img src="image.jpg" alt="Image">'
    )
    assert result
    assert result[1] == general.AltAttrPlaceable(['alt="Image"'])


def test_placeable_qt_formatting() -> None:
    result = general.QtFormattingPlaceable.parse("One %1 %99 %L1 are all valid")
    assert result
    assert result[1] == general.QtFormattingPlaceable(["%1"])
    assert result[3] == general.QtFormattingPlaceable(["%99"])
    assert result[5] == general.QtFormattingPlaceable(["%L1"])


def test_placeable_camelcase() -> None:
    result = general.CamelCasePlaceable.parse("CamelCase")
    assert result
    assert result[0] == general.CamelCasePlaceable(["CamelCase"])
    result = general.CamelCasePlaceable.parse("iPod")
    assert result
    assert result[0] == general.CamelCasePlaceable(["iPod"])
    result = general.CamelCasePlaceable.parse("DokuWiki")
    assert result
    assert result[0] == general.CamelCasePlaceable(["DokuWiki"])
    result = general.CamelCasePlaceable.parse("KBabel")
    assert result
    assert result[0] == general.CamelCasePlaceable(["KBabel"])
    assert general.CamelCasePlaceable.parse("_Bug") is None
    assert general.CamelCasePlaceable.parse("NOTCAMEL") is None


def test_placeable_space() -> None:
    result = general.SpacesPlaceable.parse(" Space at start")
    assert result
    assert result[0] == general.SpacesPlaceable([" "])
    result = general.SpacesPlaceable.parse("Space at end ")
    assert result
    assert result[1] == general.SpacesPlaceable([" "])
    result = general.SpacesPlaceable.parse("Double  space")
    assert result
    assert result[1] == general.SpacesPlaceable(["  "])


def test_placeable_punctuation() -> None:
    assert (
        general.PunctuationPlaceable.parse(
            'These, are not. Special: punctuation; marks! Or are "they"?'
        )
        is None
    )
    result = general.PunctuationPlaceable.parse("Downloading…")
    assert result
    assert result[1] == general.PunctuationPlaceable(["…"])


def test_placeable_xml_entity() -> None:
    result = general.XMLEntityPlaceable.parse("&brandShortName;")
    assert result
    assert result[0] == general.XMLEntityPlaceable(["&brandShortName;"])
    result = general.XMLEntityPlaceable.parse("&#1234;")
    assert result
    assert result[0] == general.XMLEntityPlaceable(["&#1234;"])
    result = general.XMLEntityPlaceable.parse("&xDEAD;")
    assert result
    assert result[0] == general.XMLEntityPlaceable(["&xDEAD;"])


def test_placeable_xml_tag() -> None:
    result = general.XMLTagPlaceable.parse("<a>koei</a>")
    assert result
    assert result[0] == general.XMLTagPlaceable(["<a>"])
    assert result[2] == general.XMLTagPlaceable(["</a>"])
    result = general.XMLTagPlaceable.parse("<Exif.XResolution>")
    assert result
    assert result[0] == general.XMLTagPlaceable(["<Exif.XResolution>"])
    result = general.XMLTagPlaceable.parse("<tag_a>")
    assert result
    assert result[0] == general.XMLTagPlaceable(["<tag_a>"])
    result = general.XMLTagPlaceable.parse('<img src="koei.jpg" />')
    assert result
    assert result[0] == general.XMLTagPlaceable(['<img src="koei.jpg" />'])
    # We don't want this to be recognised, so we test for None - not sure if that is a stable assumption
    assert general.XMLTagPlaceable.parse("<important word>") is None
    assert general.XMLTagPlaceable.parse('<img ="koei.jpg" />') is None
    assert general.XMLTagPlaceable.parse('<img "koei.jpg" />') is None
    result = general.XMLTagPlaceable.parse('<span xml:space="preserve">')
    assert result
    assert result[0] == general.XMLTagPlaceable(['<span xml:space="preserve">'])
    result = general.XMLTagPlaceable.parse(
        '<img src="http://translate.org.za/blogs/friedel/sites/translate.org.za.blogs.friedel/files/virtaal-7f_help.png" alt="Virtaal met lêernaam-pseudovertaling" style="border: 1px dotted grey;" />'
    )
    assert result
    assert result[0] == general.XMLTagPlaceable(
        [
            '<img src="http://translate.org.za/blogs/friedel/sites/translate.org.za.blogs.friedel/files/virtaal-7f_help.png" alt="Virtaal met lêernaam-pseudovertaling" style="border: 1px dotted grey;" />'
        ]
    )
    # Bug 933
    result = general.XMLTagPlaceable.parse(
        'This entry expires in %days% days. Would you like to <a href="%href%?PHPSESSID=5d59c559cf4eb9f1d278918271fbe68a" title="Renew this Entry Now">Renew this Entry Now</a> ?'
    )
    assert result
    assert result[1] == general.XMLTagPlaceable(
        [
            '<a href="%href%?PHPSESSID=5d59c559cf4eb9f1d278918271fbe68a" title="Renew this Entry Now">'
        ]
    )
    result = general.XMLTagPlaceable.parse(
        """<span weight='bold' size='larger'>Your Google Account is locked</span>"""
    )
    assert result
    assert result[0] == general.XMLTagPlaceable(
        ["""<span weight='bold' size='larger'>"""]
    )


def test_placeable_option() -> None:
    result = general.OptionPlaceable.parse("Type --help for this help")
    assert result
    assert result[1] == general.OptionPlaceable(["--help"])
    result = general.OptionPlaceable.parse("Short -S ones also")
    assert result
    assert result[1] == general.OptionPlaceable(["-S"])


def test_placeable_file() -> None:
    result = general.FilePlaceable.parse("Store in /home/user")
    assert result
    assert result[1] == general.FilePlaceable(["/home/user"])
    result = general.FilePlaceable.parse("Store in ~/Download directory")
    assert result
    assert result[1] == general.FilePlaceable(["~/Download"])


def test_placeable_email() -> None:
    result = general.EmailPlaceable.parse("Send email to info@example.com")
    assert result
    assert result[1] == general.EmailPlaceable(["info@example.com"])
    result = general.EmailPlaceable.parse("Send email to mailto:info@example.com")
    assert result
    assert result[1] == general.EmailPlaceable(["mailto:info@example.com"])


def test_placeable_caps() -> None:
    result = general.CapsPlaceable.parse("Use the HTML page")
    assert result
    assert result[1] == general.CapsPlaceable(["HTML"])
    assert general.CapsPlaceable.parse("I am") is None
    assert general.CapsPlaceable.parse("Use the A4 paper") is None
    result = general.CapsPlaceable.parse("In GTK+")
    assert result
    assert result[1] == general.CapsPlaceable(["GTK+"])
    #    assert general.CapsPlaceable.parse('GNOME-stuff')[0] == general.CapsPlaceable(['GNOME'])
    result = general.CapsPlaceable.parse("with XDG_USER_DIRS")
    assert result
    assert result[1] == general.CapsPlaceable(["XDG_USER_DIRS"])


def test_placeable_formatting() -> None:
    fp = general.FormattingPlaceable
    result = fp.parse("There were %d cows")
    assert result
    assert result[1] == fp(["%d"])
    result = fp.parse("There were %Id cows")
    assert result
    assert result[1] == fp(["%Id"])
    result = fp.parse("There were %d %s")
    assert result
    assert result[3] == fp(["%s"])
    result = fp.parse("%1$s was kicked by %2$s")
    assert result
    assert result[0] == fp(["%1$s"])
    result = fp.parse("There were %Id cows")
    assert result
    assert result[1] == fp(["%Id"])
    result = fp.parse("There were % d cows")
    assert result
    assert result[1] == fp(["% d"])
    # only a real space is allowed as formatting flag
    assert fp.parse("There were %\u00a0d cows") is None
    result = fp.parse("There were %'f cows")
    assert result
    assert result[1] == fp(["%'f"])
    result = fp.parse("There were %#x cows")
    assert result
    assert result[1] == fp(["%#x"])

    # field width
    result = fp.parse("There were %3d cows")
    assert result
    assert result[1] == fp(["%3d"])
    result = fp.parse("There were %33d cows")
    assert result
    assert result[1] == fp(["%33d"])
    result = fp.parse("There were %*d cows")
    assert result
    assert result[1] == fp(["%*d"])

    # numbered variables
    result = fp.parse("There were %1$d cows")
    assert result
    assert result[1] == fp(["%1$d"])


def test_placeable_doubleat() -> None:
    dap = general.DoubleAtPlaceable
    result = dap.parse("There were @@number@@ cows")
    assert result
    assert result[1] == dap(["@@number@@"])
    result = dap.parse("There were @@number1@@ cows and @@number2@@ sheep")
    assert result
    assert result[1] == dap(["@@number1@@"])
    assert result[3] == dap(["@@number2@@"])


def test_placeable_brace() -> None:
    bp = general.BracePlaceable
    # Double braces
    result = bp.parse("There were {{number}} cows")
    assert result
    assert result[1] == bp(["{{number}}"])
    result = bp.parse("There were {{number1}} cows and {{number2}} sheep")
    assert result
    assert result[1] == bp(["{{number1}}"])
    assert result[3] == bp(["{{number2}}"])

    # Single braces
    result = bp.parse("There were {number} cows")
    assert result
    assert result[1] == bp(["{number}"])
    result = bp.parse("There were {number1} cows and {number2} sheep")
    assert result
    assert result[1] == bp(["{number1}"])
    assert result[3] == bp(["{number2}"])

    # Mixed single and double braces
    result = bp.parse("There were {number1} cows and {{number2}} sheep")
    assert result
    assert result[1] == bp(["{number1}"])
    assert result[3] == bp(["{{number2}}"])
    result = bp.parse("There were {{number1}} cows and {number2} sheep")
    assert result
    assert result[1] == bp(["{{number1}}"])
    assert result[3] == bp(["{number2}"])


def test_python_placeable() -> None:
    pfp = general.PythonFormattingPlaceable
    # No conversion
    result = pfp.parse("100%% correct")
    assert result
    assert result[1] == pfp(["%%"])

    # Mapping keys
    result = pfp.parse("There were %(number)d cows")
    assert result
    assert result[1] == pfp(["%(number)d"])
    result = pfp.parse("There were %(cows.number)d cows")
    assert result
    assert result[1] == pfp(["%(cows.number)d"])
    result = pfp.parse("There were %(number of cows)d cows")
    assert result
    assert result[1] == pfp(["%(number of cows)d"])

    # Conversion flags
    result = pfp.parse("There were %(number)03d cows")
    assert result
    assert result[1] == pfp(["%(number)03d"])
    result = pfp.parse("There were %(number) 3d cows")
    assert result
    assert result[1] == pfp(["%(number) 3d"])

    # Minimum field width
    result = pfp.parse("There were %(number)*d cows")
    assert result
    assert result[1] == pfp(["%(number)*d"])

    # Precision
    result = pfp.parse("There were %(number)3.1d cows")
    assert result
    assert result[1] == pfp(["%(number)3.1d"])

    # Length modifier
    result = pfp.parse("There were %(number)Ld cows")
    assert result
    assert result[1] == pfp(["%(number)Ld"])


# TODO: JavaMessageFormatPlaceable, UrlPlaceable, XMLTagPlaceable



from translate.lang import factory


def test_punctranslate() -> None:
    """Tests that we can translate punctuation."""
    language = factory.getlanguage("or")
    assert language.punctranslate("") == ""
    assert language.punctranslate("Document loaded") == "Document loaded"
    assert language.punctranslate("Document loaded.") == "Document loaded।"
    assert language.punctranslate("Document loaded.\n") == "Document loaded।\n"
    assert language.punctranslate("Document loaded...") == "Document loaded..."


def test_country_code() -> None:
    """
    Tests that we get the correct one even if a country code is attached to
    a special code being a reserved word in Python (like 'or').
    """
    language = factory.getlanguage("or-IN")
    assert language.fullname == "Odia"


def test_sentences() -> None:
    """Tests basic functionality of sentence segmentation."""
    language = factory.getlanguage("or")
    sentences = language.sentences("")
    assert sentences == []

    sentences = language.sentences(
        "ଗୋଟିଏ ଚାବିକୁ ଆଲୋକପାତ କରିବା ପାଇଁ ମାଉସ ସୂଚକକୁ ତାହା ଉପରକୁ ଘୁଞ୍ଚାନ୍ତୁ। ଚୟନ କରିବା ପାଇଁ ଗୋଟିଏ ସୁଇଚକୁ ଦବାନ୍ତୁ।"
    )
    assert sentences == [
        "ଗୋଟିଏ ଚାବିକୁ ଆଲୋକପାତ କରିବା ପାଇଁ ମାଉସ ସୂଚକକୁ ତାହା ଉପରକୁ ଘୁଞ୍ଚାନ୍ତୁ।",
        "ଚୟନ କରିବା ପାଇଁ ଗୋଟିଏ ସୁଇଚକୁ ଦବାନ୍ତୁ।",
    ]



"""Tests for Apple XLIFF format with plural support."""

from __future__ import annotations

from translate.misc.multistring import multistring
from translate.storage import applestrings_xliff

from . import test_xliff

# ---------------------------------------------------------------------------
# Shared XLIFF snippets used across several tests
# ---------------------------------------------------------------------------

_SINGULAR_XLIFF = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" version="1.2">
  <file original="Localizable.strings" source-language="en" target-language="en" datatype="plaintext">
    <body>
      <trans-unit id="greeting" xml:space="preserve">
        <source>Hello</source>
        <target>Hello</target>
      </trans-unit>
    </body>
  </file>
</xliff>"""

_BASIC_PLURAL_XLIFF = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" version="1.2">
  <file original="Localizable.strings" source-language="en" target-language="en" datatype="plaintext">
    <body>
      <trans-unit id="items:count:dict" xml:space="preserve">
        <source>NSStringPluralRuleType</source>
        <target>NSStringPluralRuleType</target>
      </trans-unit>
      <trans-unit id="items:count:dict/:string" xml:space="preserve">
        <source>d</source>
        <target>d</target>
      </trans-unit>
      <trans-unit id="items:count:dict/one:dict/:string" xml:space="preserve">
        <source>One item</source>
        <target>One item</target>
      </trans-unit>
      <trans-unit id="items:count:dict/other:dict/:string" xml:space="preserve">
        <source>%d items</source>
        <target>%d items</target>
      </trans-unit>
    </body>
  </file>
</xliff>"""


def _add_plural(
    store: applestrings_xliff.AppleStringsXliffFile,
    base_key: str,
    target_strings: list,
    format_value_type: str = "d",
    source_strings: list | None = None,
) -> applestrings_xliff.AppleStringsXliffUnit:
    """
    Add a plural unit using the standard store API.

    :param store:              An AppleStringsXliffFile instance.
    :param base_key:           Logical key (e.g. ``"items:count"``).
    :param target_strings:     List/multistring of target plural forms.
    :param format_value_type:  NSStringFormatValueTypeKey value (e.g. ``"d"``).
    :param source_strings:     Optional source forms; defaults to *target_strings*.
    :returns:                  The created :class:`AppleStringsXliffUnit`.
    """
    if source_strings is None:
        source_strings = target_strings
    filenames = store.getfilenames()
    filename = filenames[0] if filenames else "NoName"
    unit = store.addsourceunit(
        multistring(list(source_strings)), filename=filename, createifmissing=True
    )
    unit.target = multistring(list(target_strings))
    unit.setid(base_key)
    unit.format_value_type = format_value_type
    return unit


# ---------------------------------------------------------------------------
# Unit-level tests
# ---------------------------------------------------------------------------


class TestAppleStringsXliffUnit(test_xliff.TestXLIFFUnit):
    """Tests for AppleStringsXliffUnit."""

    UnitClass = applestrings_xliff.AppleStringsXliffUnit


# ---------------------------------------------------------------------------
# File-level tests
# ---------------------------------------------------------------------------


class TestAppleStringsXliffFile(test_xliff.TestXLIFFfile):
    """Tests for AppleStringsXliffFile."""

    StoreClass = applestrings_xliff.AppleStringsXliffFile

    # ------------------------------------------------------------------
    # Basic parsing
    # ------------------------------------------------------------------

    def test_parse_basic_string(self):
        """A basic non-plural string is parsed as a single regular unit."""
        store = self.StoreClass()
        store.parse(_SINGULAR_XLIFF)

        assert len(store.units) == 1
        unit = store.units[0]
        assert unit.source == "Hello"
        assert unit.target == "Hello"
        assert not unit.hasplural()

    def test_parse_apple_plural_basic(self):
        """Plurals are folded into one unit with a multistring source/target."""
        store = self.StoreClass()
        store.parse(_BASIC_PLURAL_XLIFF)

        # The four raw trans-units are folded into a single plural unit
        assert len(store.units) == 1

        unit = store.units[0]
        assert unit.hasplural()
        assert unit.format_value_type == "d"
        assert unit._plural_base_key == "items:count"

        # source and target are multistrings (zero, one, other for English)
        assert isinstance(unit.source, multistring)
        assert isinstance(unit.target, multistring)
        assert unit.source.strings[1] == "One item"  # 'one' form
        assert unit.source.strings[2] == "%d items"  # 'other' form
        assert unit.target.strings[1] == "One item"
        assert unit.target.strings[2] == "%d items"

    def test_parse_apple_plural_complex(self):
        """Each plural group becomes one unit; regular units are preserved."""
        xliff_content = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" version="1.2">
  <file original="Localizable.strings" source-language="en" target-language="en" datatype="plaintext">
    <body>
      <trans-unit id="shopping-list" xml:space="preserve">
        <source>%1$#@apple@ and %2$#@orange@.</source>
        <target>%1$#@apple@ and %2$#@orange@.</target>
      </trans-unit>
      <trans-unit id="shopping-list:apple:dict" xml:space="preserve">
        <source>NSStringPluralRuleType</source>
        <target>NSStringPluralRuleType</target>
      </trans-unit>
      <trans-unit id="shopping-list:apple:dict/:string" xml:space="preserve">
        <source>d</source>
        <target>d</target>
      </trans-unit>
      <trans-unit id="shopping-list:apple:dict/one:dict/:string" xml:space="preserve">
        <source>One apple</source>
        <target>One apple</target>
      </trans-unit>
      <trans-unit id="shopping-list:apple:dict/other:dict/:string" xml:space="preserve">
        <source>%d apples</source>
        <target>%d apples</target>
      </trans-unit>
      <trans-unit id="shopping-list:orange:dict" xml:space="preserve">
        <source>NSStringPluralRuleType</source>
        <target>NSStringPluralRuleType</target>
      </trans-unit>
      <trans-unit id="shopping-list:orange:dict/:string" xml:space="preserve">
        <source>d</source>
        <target>d</target>
      </trans-unit>
      <trans-unit id="shopping-list:orange:dict/zero:dict/:string" xml:space="preserve">
        <source>no oranges</source>
        <target>no oranges</target>
      </trans-unit>
      <trans-unit id="shopping-list:orange:dict/one:dict/:string" xml:space="preserve">
        <source>one orange</source>
        <target>one orange</target>
      </trans-unit>
      <trans-unit id="shopping-list:orange:dict/other:dict/:string" xml:space="preserve">
        <source>%d oranges</source>
        <target>%d oranges</target>
      </trans-unit>
    </body>
  </file>
</xliff>"""
        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(xliff_content)

        # 1 regular string + 2 plural groups = 3 units
        assert len(store.units) == 3

        # The regular format string is kept as-is
        assert store.units[0].source == "%1$#@apple@ and %2$#@orange@."
        assert not store.units[0].hasplural()

        # apple group
        apple = store.units[1]
        assert apple.hasplural()
        assert apple._plural_base_key == "shopping-list:apple"
        assert apple.format_value_type == "d"
        assert apple.target.strings[1] == "One apple"  # 'one'
        assert apple.target.strings[2] == "%d apples"  # 'other'

        # orange group
        orange = store.units[2]
        assert orange.hasplural()
        assert orange._plural_base_key == "shopping-list:orange"
        assert orange.target.strings[0] == "no oranges"  # 'zero'
        assert orange.target.strings[1] == "one orange"  # 'one'
        assert orange.target.strings[2] == "%d oranges"  # 'other'

    # ------------------------------------------------------------------
    # getid / setid
    # ------------------------------------------------------------------

    def test_plural_unit_getid(self):
        """getid() on a parsed plural unit returns filename + ID_SEPARATOR + base_key."""
        store = self.StoreClass()
        store.parse(_BASIC_PLURAL_XLIFF)

        unit = store.units[0]
        assert unit.hasplural()
        # The ID should end with the base key (not the XML ":dict" suffix)
        assert unit.getid().endswith("items:count")
        assert unit.getid() == "Localizable.strings\x04items:count"

    def test_plural_unit_setid(self):
        """setid() on a plural unit updates the base key, XML id, and marks dirty."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        unit = _add_plural(store, "items:count", ["", "One item", "%d items"])

        # Initial state
        assert unit._plural_base_key == "items:count"
        assert unit.xmlelement.get("id") == "items:count:dict"
        assert unit.getid().endswith("items:count")

        # Change the logical ID
        unit.setid("widgets:count")
        assert unit._plural_base_key == "widgets:count"
        assert unit.xmlelement.get("id") == "widgets:count:dict"
        assert unit.getid().endswith("widgets:count")
        assert unit._plural_dirty

    def test_regular_unit_setid(self):
        """setid() on a regular unit behaves like the parent class."""
        store = self.StoreClass()
        store.parse(_SINGULAR_XLIFF)

        unit = store.units[0]
        assert not unit.hasplural()
        unit.setid("farewell")
        assert unit.xmlelement.get("id") == "farewell"
        assert unit.getid().endswith("farewell")

    # ------------------------------------------------------------------
    # Adding plural units using the standard API
    # ------------------------------------------------------------------

    def test_add_plural_unit_standard_api(self):
        """Plural units are added via addsourceunit + multistring + setid."""
        store = self.StoreClass()
        store.settargetlanguage("en")

        unit = _add_plural(store, "items:count", ["", "One item", "%d items"])

        assert len(store.units) == 1
        assert unit.hasplural()
        assert unit._plural_base_key == "items:count"
        assert unit.format_value_type == "d"
        assert unit.target.strings[1] == "One item"
        assert unit.target.strings[2] == "%d items"

    # ------------------------------------------------------------------
    # Serialisation
    # ------------------------------------------------------------------

    def test_serialize_plural_roundtrip(self):
        """Parse → serialize → parse produces identical plural content."""
        store = self.StoreClass()
        store.parse(_BASIC_PLURAL_XLIFF)
        output = bytes(store)

        store2 = self.StoreClass()
        store2.parse(output)

        assert len(store2.units) == 1
        unit = store2.units[0]
        assert unit.hasplural()
        assert unit.target.strings[1] == "One item"
        assert unit.target.strings[2] == "%d items"

    def test_serialize_plural_xml_structure(self):
        """The serialised XML contains the Apple XLIFF plural trans-unit structure."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        _add_plural(store, "items:count", ["", "One item", "%d items"])

        output = bytes(store).decode("utf-8")

        assert 'id="items:count:dict"' in output
        assert "<source>NSStringPluralRuleType</source>" in output

        assert 'id="items:count:dict/:string"' in output
        assert "<source>d</source>" in output

        assert 'id="items:count:dict/one:dict/:string"' in output
        assert "<source>One item</source>" in output

        assert 'id="items:count:dict/other:dict/:string"' in output
        assert "<source>%d items</source>" in output

        # zero form was empty → should be absent
        assert 'id="items:count:dict/zero:dict/:string"' not in output

    def test_serialize_plural_all_forms(self):
        """All three English plural forms are serialised when provided."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        _add_plural(store, "items:count", ["No items", "One item", "%d items"])

        output = bytes(store).decode("utf-8")

        assert 'id="items:count:dict/zero:dict/:string"' in output
        assert 'id="items:count:dict/one:dict/:string"' in output
        assert 'id="items:count:dict/other:dict/:string"' in output
        assert "<source>No items</source>" in output
        assert "<source>One item</source>" in output
        assert "<source>%d items</source>" in output

    # ------------------------------------------------------------------
    # Conversion: singular → plural
    # ------------------------------------------------------------------

    def test_singular_to_plural_by_multistring_assignment(self):
        """A regular unit is converted to a plural unit by assigning a multistring."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(_SINGULAR_XLIFF)

        unit = store.units[0]
        assert not unit.hasplural()
        assert unit.source == "Hello"

        # Convert to plural
        unit.source = multistring(["", "One item", "%d items"])
        unit.target = multistring(["", "Jeden Artikel", "%d Artikel"])
        unit.setid("items:count")
        unit.format_value_type = "d"

        assert unit.hasplural()
        assert unit._plural_base_key == "items:count"
        assert unit.target.strings[1] == "Jeden Artikel"
        assert unit.target.strings[2] == "%d Artikel"

        # Serialise – the XML must have the full Apple XLIFF plural structure
        output = bytes(store).decode("utf-8")
        assert 'id="items:count:dict"' in output
        assert 'id="items:count:dict/one:dict/:string"' in output
        assert "Jeden Artikel" in output

        # Round-trip
        store2 = self.StoreClass()
        store2.settargetlanguage("en")
        store2.parse(output.encode("utf-8"))
        assert len(store2.units) == 1
        assert store2.units[0].hasplural()
        assert store2.units[0].target.strings[1] == "Jeden Artikel"

    # ------------------------------------------------------------------
    # Conversion: plural → singular
    # ------------------------------------------------------------------

    def test_plural_to_singular_by_string_assignment(self):
        """Assigning a plain string clears the plural state of a unit."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(_BASIC_PLURAL_XLIFF)

        unit = store.units[0]
        assert unit.hasplural()

        # Convert to singular
        unit.source = "Item count"
        unit.target = "Anzahl der Objekte"

        assert not unit.hasplural()
        assert unit.source == "Item count"
        assert unit.target == "Anzahl der Objekte"

    def test_plural_to_singular_full_roundtrip(self):
        """A plural-to-singular conversion serialises cleanly."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(_BASIC_PLURAL_XLIFF)

        unit = store.units[0]
        assert unit.hasplural()

        # Convert to singular: clear plural state then fix the id
        unit.source = "Item count"
        unit.target = "Anzahl"
        unit.setid("items_label")  # clears _plural_base_key, removes stale elements

        output = bytes(store).decode("utf-8")
        assert 'id="items_label"' in output
        assert "<source>Item count</source>" in output
        assert "Anzahl" in output
        # stale plural XML elements must be gone
        assert "items:count:dict/:string" not in output
        assert "one:dict/:string" not in output

        # Round-trip
        store2 = self.StoreClass()
        store2.parse(output.encode("utf-8"))
        assert len(store2.units) == 1
        assert not store2.units[0].hasplural()
        assert store2.units[0].target == "Anzahl"

    # ------------------------------------------------------------------
    # remove_plural_unit / removeunit
    # ------------------------------------------------------------------

    def test_remove_plural_unit(self):
        """removeunit() on a plural unit removes it and its sibling XML elements."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        _add_plural(store, "items:count", ["", "One item", "%d items"])
        _add_plural(store, "orders:count", ["", "One order", "%d orders"])

        assert len(store.units) == 2

        # Find and remove the items:count plural unit
        items_unit = next(u for u in store.units if u._plural_base_key == "items:count")
        store.removeunit(items_unit)
        assert len(store.units) == 1

        # Remaining unit is for orders:count
        remaining = store.units[0]
        assert remaining._plural_base_key == "orders:count"
        assert remaining.target.strings[1] == "One order"

    def test_removeunit_cleans_sibling_xml(self):
        """removeunit() on a plural unit also removes the sibling XML elements."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(_BASIC_PLURAL_XLIFF)

        unit = store.units[0]
        assert unit.hasplural()
        store.removeunit(unit)

        assert len(store.units) == 0
        output = bytes(store).decode("utf-8")
        assert "items:count" not in output

    # ------------------------------------------------------------------
    # Mixed store: singular + plural
    # ------------------------------------------------------------------

    def test_convert_non_plural_to_plural(self):
        """A non-plural unit is removed and replaced by a plural group."""
        xliff_content = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" version="1.2">
  <file original="Localizable.strings" source-language="en" target-language="en" datatype="plaintext">
    <body>
      <trans-unit id="greeting" xml:space="preserve">
        <source>Hello</source>
        <target>Hello</target>
      </trans-unit>
      <trans-unit id="item_count" xml:space="preserve">
        <source>%d items</source>
        <target>%d items</target>
      </trans-unit>
    </body>
  </file>
</xliff>"""

        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(xliff_content)

        assert len(store.units) == 2

        # Remove the non-plural unit
        unit_to_replace = next(
            u for u in store.units if u.xmlelement.get("id") == "item_count"
        )
        store.removeunit(unit_to_replace)
        assert len(store.units) == 1

        # Add a plural unit in its place (1 original + 1 plural = 2 units)
        _add_plural(
            store,
            "item_count:count",
            ["", "One item", "%d items"],
            source_strings=["", "One item", "%d items"],
        )
        assert len(store.units) == 2

        output = bytes(store).decode("utf-8")
        assert 'id="greeting"' in output
        assert 'id="item_count:count:dict"' in output
        assert 'id="item_count:count:dict/one:dict/:string"' in output
        assert 'id="item_count:count:dict/other:dict/:string"' in output

        # Round-trip
        store2 = self.StoreClass()
        store2.settargetlanguage("en")
        store2.parse(output.encode("utf-8"))
        assert len(store2.units) == 2

        plural = next(u for u in store2.units if u.hasplural())
        assert plural._plural_base_key == "item_count:count"
        assert plural.target.strings[1] == "One item"
        assert plural.target.strings[2] == "%d items"

    def test_convert_plural_to_non_plural(self):
        """A plural group is removed and replaced by a regular unit."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(_BASIC_PLURAL_XLIFF)

        # After folding: 1 merged plural = 1 unit
        assert len(store.units) == 1

        plural_unit = store.units[0]
        assert plural_unit._plural_base_key == "items:count"
        store.removeunit(plural_unit)
        assert len(store.units) == 0

        new_unit = store.addsourceunit(
            "Items", filename="Localizable.strings", createifmissing=True
        )
        new_unit.setid("items_label")
        new_unit.target = "Items"
        assert len(store.units) == 1

        output = bytes(store).decode("utf-8")
        assert 'id="items_label"' in output
        assert "items:count:dict" not in output

        # Round-trip
        store2 = self.StoreClass()
        store2.parse(output.encode("utf-8"))
        assert len(store2.units) == 1
        assert store2.units[0].target == "Items"

    def test_add_plural_unit_to_parsed_store(self):
        """A plural unit can be added to an already-parsed store."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(_SINGULAR_XLIFF)
        assert len(store.units) == 1

        _add_plural(store, "items:count", ["", "One item", "%d items"])
        assert len(store.units) == 2

        plural = next(u for u in store.units if u.hasplural())
        assert plural.target.strings[1] == "One item"
        assert plural.target.strings[2] == "%d items"

    # ------------------------------------------------------------------
    # Target update round-trip
    # ------------------------------------------------------------------

    def test_plural_unit_target_update_roundtrip(self):
        """Changing a parsed plural unit's target round-trips correctly."""
        store = self.StoreClass()
        store.settargetlanguage("en")
        store.parse(_BASIC_PLURAL_XLIFF)

        unit = store.units[0]
        assert unit.hasplural()

        unit.target = multistring(["", "Jeden Artikel", "%d Artikel"])

        output = bytes(store).decode("utf-8")
        assert "Jeden Artikel" in output
        assert "%d Artikel" in output
        assert 'id="items:count:dict/one:dict/:string"' in output
        assert 'id="items:count:dict/other:dict/:string"' in output

        store2 = self.StoreClass()
        store2.settargetlanguage("en")
        store2.parse(output.encode("utf-8"))
        assert len(store2.units) == 1
        assert store2.units[0].target.strings[1] == "Jeden Artikel"
        assert store2.units[0].target.strings[2] == "%d Artikel"

    # ------------------------------------------------------------------
    # hasplural + getid (combined)
    # ------------------------------------------------------------------

    def test_plural_unit_hasplural_and_getid(self):
        """hasplural() and getid() behave correctly for both unit types."""
        xliff_content = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" version="1.2">
  <file original="L.strings" source-language="en" target-language="en" datatype="plaintext">
    <body>
      <trans-unit id="hello" xml:space="preserve">
        <source>Hello</source><target>Hello</target>
      </trans-unit>
      <trans-unit id="n:count:dict" xml:space="preserve">
        <source>NSStringPluralRuleType</source><target>NSStringPluralRuleType</target>
      </trans-unit>
      <trans-unit id="n:count:dict/:string" xml:space="preserve">
        <source>d</source><target>d</target>
      </trans-unit>
      <trans-unit id="n:count:dict/one:dict/:string" xml:space="preserve">
        <source>one</source><target>one</target>
      </trans-unit>
      <trans-unit id="n:count:dict/other:dict/:string" xml:space="preserve">
        <source>other</source><target>other</target>
      </trans-unit>
    </body>
  </file>
</xliff>"""
        store = self.StoreClass()
        store.parse(xliff_content)

        assert len(store.units) == 2

        regular = store.units[0]
        assert not regular.hasplural()
        assert "hello" in regular.getid()

        plural = store.units[1]
        assert plural.hasplural()
        # getid() returns the base key WITHOUT the ":dict" XML suffix
        assert plural.getid().endswith("n:count")
        assert "n:count:dict" not in plural.getid()

    # ------------------------------------------------------------------
    # Language detection
    # ------------------------------------------------------------------

    def test_get_base_key(self):
        """get_base_key() extracts the correct base key from plural-style XML ids."""
        xliff_content = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" version="1.2">
  <file original="test" source-language="en" target-language="en" datatype="plaintext">
    <body>
      <trans-unit id="key:var:dict" xml:space="preserve">
        <source>NSStringPluralRuleType</source>
      </trans-unit>
    </body>
  </file>
</xliff>"""
        store = self.StoreClass()
        store.parse(xliff_content)
        unit = store.units[0]
        assert unit.get_base_key() == "key:var"

    def test_targetlanguage_auto_detection_filename(self):
        """Target language is auto-detected from the .lproj directory name."""
        store = self.StoreClass()
        store.filename = "Project/it.lproj/Localizable.xliff"
        assert store.gettargetlanguage() == "it"

    def test_targetlanguage_auto_detection_base_filename(self):
        """Base.lproj is treated as English."""
        store = self.StoreClass()
        store.filename = "Project/Base.lproj/Localizable.xliff"
        assert store.gettargetlanguage() == "en"

    # ------------------------------------------------------------------
    # Real-world patterns (ONLYOFFICE-style marker-less plurals)
    # ------------------------------------------------------------------

    def test_real_world_plural_patterns(self):
        """Marker-less real-world plural patterns (ONLYOFFICE style) are handled."""
        xliff_content = b"""<?xml version="1.0" encoding="UTF-8"?>
<xliff xmlns="urn:oasis:names:tc:xliff:document:1.2" version="1.2">
  <file original="Localizable.strings" source-language="en" target-language="lt" datatype="plaintext">
    <body>
      <trans-unit id="/%d days are left until the license expiration.:dict/NSStringLocalizedFormatKey:dict/:string">
        <source>%#@days@</source>
        <target>%#@dienos@</target>
      </trans-unit>
      <trans-unit id="/%d days are left until the license expiration.:dict/days:dict/few:dict/:string">
        <source>%d days are left until the license expiration.</source>
        <target state="translated">Days left until expiration (few).</target>
      </trans-unit>
      <trans-unit id="/%d days are left until the license expiration.:dict/days:dict/many:dict/:string">
        <source>%d days are left until the license expiration.</source>
        <target state="translated">Days left until expiration (many).</target>
      </trans-unit>
      <trans-unit id="/%d days are left until the license expiration.:dict/days:dict/one:dict/:string">
        <source>%d day are left until the license expiration.</source>
        <target state="translated">Days left until expiration (one).</target>
      </trans-unit>
      <trans-unit id="/%d days are left until the license expiration.:dict/days:dict/other:dict/:string">
        <source>%d days are left until the license expiration.</source>
        <target state="translated">Days left until expiration (other).</target>
      </trans-unit>
    </body>
  </file>
</xliff>"""

        store = self.StoreClass()
        store.settargetlanguage("lt")
        store.parse(xliff_content)

        # NSStringLocalizedFormatKey (format_type only, no forms) → regular unit
        # days plural group → folded unit
        assert len(store.units) == 2

        format_key_unit = store.units[0]
        assert "NSStringLocalizedFormatKey" in format_key_unit.xmlelement.get("id")
        assert not format_key_unit.hasplural()

        days_unit = store.units[1]
        assert days_unit.hasplural()
        # Lithuanian has multiple plural forms (few, many, one, other)
        assert len([s for s in days_unit.target.strings if s]) >= 4



from io import BytesIO

from translate.convert import po2ts
from translate.storage import po

from . import test_convert


class TestPO2TS:
    @staticmethod
    def po2ts(posource):
        """Helper that converts po source to ts source without requiring files."""
        inputfile = BytesIO(posource.encode())
        inputpo = po.pofile(inputfile)
        convertor = po2ts.po2ts()
        output = BytesIO()
        convertor.convertstore(inputpo, output)
        return output.getvalue().decode("utf-8")

    @staticmethod
    def singleelement(storage):
        """Checks that the pofile contains a single non-header element, and returns it."""
        assert len(storage.units) == 1
        return storage.units[0]

    def test_simpleunit(self) -> None:
        """Checks that a simple po entry definition converts properly to a ts entry."""
        minipo = r'''#: term.cpp
msgid "Term"
msgstr "asdf"'''
        tsfile = self.po2ts(minipo)
        print(tsfile)
        assert "<name>term.cpp</name>" in tsfile
        assert "<source>Term</source>" in tsfile
        assert "<translation>asdf</translation>" in tsfile
        assert "<comment>" not in tsfile

    def test_simple_unicode_unit(self) -> None:
        """Checks that a simple unit with unicode strings."""
        minipo = r'''#: unicode.cpp
msgid "ßource"
msgstr "†arget"'''
        tsfile = self.po2ts(minipo)
        print(tsfile)
        print(type(tsfile))
        assert "<name>unicode.cpp</name>" in tsfile
        assert "<source>ßource</source>" in tsfile
        assert "<translation>†arget</translation>" in tsfile

    def test_fullunit(self) -> None:
        """Check that an entry with various settings is converted correctly."""
        posource = """# Translator comment
#. Automatic comment
#: location.cpp:100
msgid "Source"
msgstr "Target"
"""
        tsfile = self.po2ts(posource)
        print(tsfile)
        # The other section are a duplicate of test_simplentry
        # FIXME need to think about auto vs trans comments maybe in TS v1.1
        assert "<comment>Translator comment</comment>" in tsfile

    def test_fuzzyunit(self) -> None:
        """Check that we handle fuzzy units correctly."""
        posource = '''#: term.cpp
#, fuzzy
msgid "Source"
msgstr "Target"'''
        tsfile = self.po2ts(posource)
        print(tsfile)
        assert """<translation type="unfinished">Target</translation>""" in tsfile

    def test_obsolete(self) -> None:
        """Test that we can take back obsolete messages."""
        posource = '''#. (obsolete)
#: term.cpp
msgid "Source"
msgstr "Target"'''
        tsfile = self.po2ts(posource)
        print(tsfile)
        assert """<translation type="obsolete">Target</translation>""" in tsfile

    def test_duplicates(self) -> None:
        """Test that we can handle duplicates in the same context block."""
        posource = """#: @@@#1
msgid "English"
msgstr "a"

#: @@@#3
msgid "English"
msgstr "b"
"""
        tsfile = self.po2ts(posource)
        print(tsfile)
        assert tsfile.find("English") != tsfile.rfind("English")

    def test_linebreak(self) -> None:
        """Test that we can handle linebreaks."""
        minipo = r'''#: linebreak.cpp
msgid "Line 1\n"
"Line 2"
msgstr "Linea 1\n"
"Linea 2"'''
        tsfile = self.po2ts(minipo)
        print(tsfile)
        print(type(tsfile))
        assert "<name>linebreak.cpp</name>" in tsfile
        assert (
            r"""<source>Line 1
Line 2</source>"""
            in tsfile
        )
        assert (
            r"""<translation>Linea 1
Linea 2</translation>"""
            in tsfile
        )

    def test_linebreak_consecutive(self) -> None:
        """Test that we can handle consecutive linebreaks."""
        minipo = r'''#: linebreak.cpp
msgid "Line 1\n"
"\n"
"Line 3"
msgstr "Linea 1\n"
"\n"
"Linea 3"'''
        tsfile = self.po2ts(minipo)
        print(tsfile)
        print(type(tsfile))
        assert "<name>linebreak.cpp</name>" in tsfile
        assert (
            r"""<source>Line 1

Line 3</source>"""
            in tsfile
        )
        assert (
            r"""<translation>Linea 1

Linea 3</translation>"""
            in tsfile
        )

    def test_plural(self) -> None:
        """Test that plural forms are correctly converted from PO to TS."""
        minipo = r'''#: ExamplePO
msgid "%d day ago"
msgid_plural "%d days ago"
msgstr[0] "%d dźeń tamu"
msgstr[1] "%d dni tamu"
msgstr[2] "%d dźon tamu"'''
        tsfile = self.po2ts(minipo)
        print(tsfile)
        assert "<name>ExamplePO</name>" in tsfile
        # The source should merge singular and plural with (s) notation
        assert "<source>%d day(s) ago</source>" in tsfile
        assert 'numerus="yes"' in tsfile
        assert "<numerusform>%d dźeń tamu</numerusform>" in tsfile
        assert "<numerusform>%d dni tamu</numerusform>" in tsfile
        assert "<numerusform>%d dźon tamu</numerusform>" in tsfile

    def test_plural_fuzzy(self) -> None:
        """Test that fuzzy plural forms are handled correctly."""
        minipo = r'''#: FuzzyContext
#, fuzzy
msgid "%d item"
msgid_plural "%d items"
msgstr[0] "%d punkt"
msgstr[1] "%d punkty"'''
        tsfile = self.po2ts(minipo)
        print(tsfile)
        assert "<name>FuzzyContext</name>" in tsfile
        assert "<source>%d item(s)</source>" in tsfile
        assert 'numerus="yes"' in tsfile
        assert 'type="unfinished"' in tsfile
        assert "<numerusform>%d punkt</numerusform>" in tsfile
        assert "<numerusform>%d punkty</numerusform>" in tsfile

    def test_unit_without_location(self) -> None:
        """Test that units without location comments are still converted."""
        # This is a regression test for the issue where strings without
        # location comments (no #: line) were not being converted to TS
        minipo = r'''msgid "File Path"
msgstr "Ruta del archivo"'''
        tsfile = self.po2ts(minipo)
        print(tsfile)
        # The unit should be present even without a location
        assert "<source>File Path</source>" in tsfile
        assert "<translation>Ruta del archivo</translation>" in tsfile
        # It should be in a context with an empty name
        assert "<name></name>" in tsfile

    def test_mixed_units_with_and_without_location(self) -> None:
        """Test conversion of both units with and without location comments."""
        minipo = r'''#: mainWindow#1
msgid "Machine"
msgstr "Máquina"

msgid "File Path"
msgstr "Ruta del archivo"'''
        tsfile = self.po2ts(minipo)
        print(tsfile)
        # Both units should be present
        assert "<source>Machine</source>" in tsfile
        assert "<translation>Máquina</translation>" in tsfile
        assert "<source>File Path</source>" in tsfile
        assert "<translation>Ruta del archivo</translation>" in tsfile
        # First should have mainWindow context
        assert "<name>mainWindow</name>" in tsfile
        # Second should have empty context
        assert "<name></name>" in tsfile


class TestPO2TSCommand(test_convert.TestConvertCommand, TestPO2TS):
    """Tests running actual po2ts commands on files."""

    convertmodule = po2ts

    expected_options = [
        "-c CONTEXT, --context=CONTEXT",
        "-t TEMPLATE, --template=TEMPLATE",
    ]



from io import BytesIO

from translate.convert import csv2po, po2csv
from translate.storage import csvl10n, po

from ..storage.test_base import first_translatable, headerless_len
from . import test_convert


class TestPO2CSV:
    @staticmethod
    def po2csv(posource):
        """Helper that converts po source to csv source without requiring files."""
        inputfile = BytesIO(posource.encode())
        inputpo = po.pofile(inputfile)
        convertor = po2csv.po2csv()
        return convertor.convertstore(inputpo)

    @staticmethod
    def csv2po(csvsource, template=None):
        """Helper that converts csv source to po source without requiring files."""
        inputfile = BytesIO(csvsource)
        inputcsv = csvl10n.csvfile(inputfile)
        if template:
            templatefile = BytesIO(template.encode())
            inputpot = po.pofile(templatefile)
        else:
            inputpot = None
        convertor = csv2po.csv2po(templatepo=inputpot)
        return convertor.convertstore(inputcsv)

    @staticmethod
    def singleelement(storage):
        """Checks that the pofile contains a single non-header element, and returns it."""
        assert headerless_len(storage.units) == 1
        return first_translatable(storage)

    def test_simpleentity(self) -> None:
        """Checks that a simple csv entry definition converts properly to a po entry."""
        minipo = r'''#: term.cpp
msgid "Term"
msgstr "asdf"'''
        csvfile = self.po2csv(minipo)
        unit = self.singleelement(csvfile)
        assert unit.location == "term.cpp"
        assert unit.source == "Term"
        assert unit.target == "asdf"

    def test_multiline(self) -> None:
        """Tests multiline po entries."""
        minipo = r'''msgid "First part "
"and extra"
msgstr "Eerste deel "
"en ekstra"'''
        csvfile = self.po2csv(minipo)
        unit = self.singleelement(csvfile)
        assert unit.source == "First part and extra"
        assert unit.target == "Eerste deel en ekstra"

    def test_escapednewlines(self) -> None:
        """Test the escaping of newlines."""
        minipo = r"""msgid "First line\nSecond line"
msgstr "Eerste lyn\nTweede lyn"
"""
        csvfile = self.po2csv(minipo)
        unit = self.singleelement(csvfile)
        assert unit.source == "First line\nSecond line"
        assert unit.target == "Eerste lyn\nTweede lyn"
        pofile = self.csv2po(bytes(csvfile))
        unit = self.singleelement(pofile)
        assert unit.source == "First line\nSecond line"
        assert unit.target == "Eerste lyn\nTweede lyn"

    def test_escapedtabs(self) -> None:
        """Test the escaping of tabs."""
        minipo = r"""msgid "First column\tSecond column"
msgstr "Eerste kolom\tTweede kolom"
"""
        csvfile = self.po2csv(minipo)
        unit = self.singleelement(csvfile)
        assert unit.source == "First column\tSecond column"
        assert unit.target == "Eerste kolom\tTweede kolom"
        assert (
            csvfile.findunit("First column\tSecond column").target
            == "Eerste kolom\tTweede kolom"
        )

    def test_escapedquotes(self) -> None:
        """Test the escaping of quotes (and slash)."""
        minipo = r"""msgid "Hello \"Everyone\""
msgstr "Good day \"All\""

msgid "Use \\\"."
msgstr "Gebruik \\\"."
"""
        csvfile = self.po2csv(minipo)
        assert csvfile.findunit('Hello "Everyone"').target == 'Good day "All"'
        assert csvfile.findunit('Use \\".').target == 'Gebruik \\".'

    def test_escapedescape(self) -> None:
        """Test the escaping of pure escapes is unaffected."""
        minipo = r"""msgid "Find\\Options"
msgstr "Vind\\Opsies"
"""
        csvfile = self.po2csv(minipo)
        print(minipo)
        print(csvfile)
        assert csvfile.findunit(r"Find\Options").target == r"Vind\Opsies"

    def test_singlequotes(self) -> None:
        """Tests that single quotes are preserved correctly."""
        minipo = """msgid "source 'source'"\nmsgstr "target 'target'"\n"""
        csvfile = self.po2csv(minipo)
        print(bytes(csvfile))
        assert csvfile.findunit("source 'source'").target == "target 'target'"
        # Make sure we don't mess with start quotes until writing
        minipo = """msgid "'source'"\nmsgstr "'target'"\n"""
        csvfile = self.po2csv(minipo)
        print(bytes(csvfile))
        assert csvfile.findunit(r"'source'").target == r"'target'"
        # TODO check that we escape on writing not in the internal representation

    def test_empties(self) -> None:
        """Tests that things keep working with empty entries."""
        minipo = 'msgid "Source"\nmsgstr ""\n\nmsgid ""\nmsgstr ""'
        csvfile = self.po2csv(minipo)
        assert csvfile.findunit("Source") is not None
        assert csvfile.findunit("Source").target == ""
        assert headerless_len(csvfile.units) == 1

    def test_kdecomments(self) -> None:
        """Test that we don't carry KDE comments to CSV."""
        minipo = '#: simple.c\nmsgid "_: KDE comment\\n"\n"Same"\nmsgstr "Same"\n'
        csvfile = self.po2csv(minipo)
        unit = self.singleelement(csvfile)
        assert unit.source == "Same"
        assert unit.target == "Same"


class TestPO2CSVCommand(test_convert.TestConvertCommand, TestPO2CSV):
    """Tests running actual po2csv commands on files."""

    convertmodule = po2csv

    expected_options = [
        "--columnorder=COLUMNORDER",
    ]

    def test_columnorder(self) -> None:
        pocontent = '#: simple.c\nmsgid "Same"\nmsgstr "Target"\n'
        self.create_testfile("test.po", pocontent)

        self.run_command("test.po", "test.csv")
        content = self.open_testfile("test.csv", "r").read()
        assert (
            content
            == """"location","source","target"
"simple.c","Same","Target"
"""
        )

        self.run_command("test.po", "test.csv", columnorder="target,source")
        content = self.open_testfile("test.csv", "r").read()
        assert (
            content
            == """"target","source"
"Target","Same"
"""
        )

    def test_context(self) -> None:
        pocontent = '#: simple.c\nmsgctxt "Context"\nmsgid "Same"\nmsgstr "Target"\n'
        self.create_testfile("test.po", pocontent)

        self.run_command("test.po", "test.csv", columnorder="context,source,target")
        content = self.open_testfile("test.csv", "r").read()
        assert (
            content
            == """"context","source","target"
"Context","Same","Target"
"""
        )



from io import BytesIO

from translate.lang import factory as lang_factory
from translate.storage import po, tmx, xliff
from translate.tools import posegment


class TestPOSegment:
    @staticmethod
    def posegment(
        posource,
        sourcelanguage,
        targetlanguage,
        stripspaces=True,
        onlyaligned=True,
    ):
        """Helper that convert po source without requiring files."""
        inputfile = BytesIO(posource.encode())
        inputpo = po.pofile(inputfile)
        sourcelang = lang_factory.getlanguage(sourcelanguage)
        targetlang = lang_factory.getlanguage(targetlanguage)
        convertor = posegment.segment(
            sourcelang, targetlang, stripspaces=stripspaces, onlyaligned=onlyaligned
        )
        return convertor.convertstore(inputpo)

    def test_en_ja_simple(self) -> None:
        posource = """
#: test/test.py:112
msgid ""
"Please let us know if you have any specific needs (A/V requirements, "
"multiple microphones, a table, etc).  Note for example that 'audio out' is "
"not provided for your computer unless you tell us in advance."
msgstr ""
"特に必要な物(A/V機器、複数のマイク、テーブルetc)があれば教えて下さい。例とし"
"て、コンピュータからの「音声出力」は事前にお知らせ頂いていない場合は提供でき"
"ないことに注意して下さい。"
"""
        poresult = self.posegment(posource, "en", "ja")
        out_unit = poresult.units[1]
        assert (
            out_unit.source
            == "Please let us know if you have any specific needs (A/V requirements, multiple microphones, a table, etc)."
        )
        assert (
            out_unit.target
            == "特に必要な物(A/V機器、複数のマイク、テーブルetc)があれば教えて下さい。"
        )
        out_unit = poresult.units[2]
        assert (
            out_unit.source
            == "Note for example that 'audio out' is not provided for your computer unless you tell us in advance."
        )
        assert (
            out_unit.target
            == "例として、コンピュータからの「音声出力」は事前にお知らせ頂いていない場合は提供できないことに注意して下さい。"
        )

    def test_en_ja_punctuation(self) -> None:
        """Checks that a half-width punctuation."""
        posource = """
#: docs/intro/contributing.txt:184
msgid ""
"Note that the latest Django trunk may not always be stable. When developing "
"against trunk, you can check Django's continuous integration builds."
msgstr ""
"開発中の､最新の Django ではステーブルとは限りません｡トランクバージョンで開発"
"を行う場合､ Django の継続インテグレーションビルドをチェックしてください｡"
"""
        poresult = self.posegment(posource, "en", "ja")
        out_unit = poresult.units[1]
        assert (
            out_unit.source
            == "Note that the latest Django trunk may not always be stable."
        )
        assert out_unit.target == "開発中の､最新の Django ではステーブルとは限りません｡"

        out_unit = poresult.units[2]
        assert (
            out_unit.source
            == "When developing against trunk, you can check Django's continuous integration builds."
        )
        assert (
            out_unit.target
            == "トランクバージョンで開発を行う場合､ Django の継続インテグレーションビルドをチェックしてください｡"
        )

    def test_transifex_po(self) -> None:
        """Checks that onlyaligned=True removes units with mismatched sentence counts."""
        # First unit: mismatched (3 EN sentences vs 4 JA sentences) - should be removed
        # Second unit: matched (2 EN sentences vs 2 JA sentences) - should be kept
        posource = """# 8456b7744a02478cb75a9ad7f950d8cf
#: ../../../../1.8/docs/intro/contributing.txt:184
msgid ""
"Note that the latest Django trunk may not always be stable. When developing "
"against trunk, you can check `Django's continuous integration builds`__ to "
"determine if the failures are specific to your machine or if they are also "
"present in Django's official builds. If you click to view a particular "
"build, you can view the Configuration Matrix which shows failures broken "
"down by Python version and database backend."
msgstr ""
"開発中の､最新の Django ではステーブルとは限りません｡トランクバージョンで開発"
"を行う場合､ `Django の継続インテグレーションビルド`__ をチェックしてください｡"
"これで､テストの失敗があなたのマシンだけのものか､ Django 公式のビルドによるも"
"のかが分かります｡各ビルドについてのリンクをクリックすれば､ Configuration "
"Matrix という､ 各 Python のバージョン､ DB バックエンドに対応したテストの失"
"敗を閲覧できます｡"

#: docs/intro/contributing.txt:200
msgid ""
"Note that the latest Django trunk may not always be stable. When developing "
"against trunk, you can check Django's continuous integration builds."
msgstr ""
"開発中の､最新の Django ではステーブルとは限りません｡トランクバージョンで開発"
"を行う場合､ Django の継続インテグレーションビルドをチェックしてください｡"
"""
        poresult = self.posegment(posource, "en", "ja")
        # With onlyaligned=True, only the aligned unit should be kept (plus header)
        # First unit (mismatched 3 vs 4) is removed, second unit (matched 2 vs 2) is segmented
        assert len(poresult.units) == 3  # header + 2 segments from second unit

        # Check the segmented units from the second (matched) unit
        out_unit = poresult.units[1]
        assert (
            out_unit.source
            == "Note that the latest Django trunk may not always be stable."
        )
        assert out_unit.target == "開発中の､最新の Django ではステーブルとは限りません｡"

        out_unit = poresult.units[2]
        assert (
            out_unit.source
            == "When developing against trunk, you can check Django's continuous integration builds."
        )
        assert (
            out_unit.target
            == "トランクバージョンで開発を行う場合､ Django の継続インテグレーションビルドをチェックしてください｡"
        )


class TestXLIFFSegment:
    @staticmethod
    def xliff_segment(
        xliffsource,
        sourcelanguage,
        targetlanguage,
        stripspaces=True,
        onlyaligned=True,
    ):
        """Helper that converts xliff source without requiring files."""
        inputfile = BytesIO(xliffsource.encode())
        inputstore = xliff.xlifffile(inputfile)
        sourcelang = lang_factory.getlanguage(sourcelanguage)
        targetlang = lang_factory.getlanguage(targetlanguage)
        convertor = posegment.segment(
            sourcelang, targetlang, stripspaces=stripspaces, onlyaligned=onlyaligned
        )
        return convertor.convertstore(inputstore)

    def test_xliff_en_simple(self) -> None:
        """Test basic XLIFF segmentation."""
        xliffsource = """<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
  <file datatype="plaintext" original="test.txt" source-language="en" target-language="fr">
    <body>
      <trans-unit id="1" approved="yes">
        <source>This is the first sentence. This is the second sentence.</source>
        <target>Ceci est la première phrase. Ceci est la deuxième phrase.</target>
      </trans-unit>
    </body>
  </file>
</xliff>"""
        xliffresult = self.xliff_segment(xliffsource, "en", "fr")
        # Should have 2 units after segmentation
        assert len(xliffresult.units) == 2
        assert xliffresult.units[0].source == "This is the first sentence."
        assert xliffresult.units[0].target == "Ceci est la première phrase."
        assert xliffresult.units[1].source == "This is the second sentence."
        assert xliffresult.units[1].target == "Ceci est la deuxième phrase."

    def test_xliff_untranslated(self) -> None:
        """Test XLIFF segmentation with untranslated units."""
        xliffsource = """<?xml version="1.0" encoding="utf-8"?>
<xliff version="1.1" xmlns="urn:oasis:names:tc:xliff:document:1.1">
  <file datatype="plaintext" original="test.txt" source-language="en" target-language="fr">
    <body>
      <trans-unit id="1">
        <source>First sentence. Second sentence.</source>
      </trans-unit>
    </body>
  </file>
</xliff>"""
        xliffresult = self.xliff_segment(xliffsource, "en", "fr")
        # Should have 2 units after segmentation with empty targets
        assert len(xliffresult.units) == 2
        assert xliffresult.units[0].source == "First sentence."
        assert xliffresult.units[0].target == ""
        assert xliffresult.units[1].source == "Second sentence."
        assert xliffresult.units[1].target == ""


class TestTMXSegment:
    @staticmethod
    def tmx_segment(
        tmxsource,
        sourcelanguage,
        targetlanguage,
        stripspaces=True,
        onlyaligned=True,
    ):
        """Helper that converts tmx source without requiring files."""
        inputfile = BytesIO(tmxsource.encode())
        inputstore = tmx.tmxfile(inputfile)
        sourcelang = lang_factory.getlanguage(sourcelanguage)
        targetlang = lang_factory.getlanguage(targetlanguage)
        convertor = posegment.segment(
            sourcelang, targetlang, stripspaces=stripspaces, onlyaligned=onlyaligned
        )
        return convertor.convertstore(inputstore)

    def test_tmx_en_simple(self) -> None:
        """Test basic TMX segmentation."""
        tmxsource = """<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE tmx SYSTEM "tmx14.dtd">
<tmx version="1.4">
  <header adminlang="en" creationtool="test" creationtoolversion="1.0"
          datatype="PlainText" o-tmf="UTF-8" segtype="sentence" srclang="en"/>
  <body>
    <tu>
      <tuv xml:lang="en">
        <seg>This is the first sentence. This is the second sentence.</seg>
      </tuv>
      <tuv xml:lang="fr">
        <seg>Ceci est la première phrase. Ceci est la deuxième phrase.</seg>
      </tuv>
    </tu>
  </body>
</tmx>"""
        tmxresult = self.tmx_segment(tmxsource, "en", "fr")
        # Should have 2 units after segmentation
        assert len(tmxresult.units) == 2
        assert tmxresult.units[0].source == "This is the first sentence."
        assert tmxresult.units[0].target == "Ceci est la première phrase."
        assert tmxresult.units[1].source == "This is the second sentence."
        assert tmxresult.units[1].target == "Ceci est la deuxième phrase."



from io import BytesIO

from translate.convert import prop2po
from translate.storage import po, properties

from . import test_convert


class TestProp2PO:
    @staticmethod
    def prop2po(propsource, proptemplate=None, personality="java"):
        """Helper that converts .properties source to po source without requiring files."""
        inputfile = BytesIO(propsource.encode())
        inputprop = properties.propfile(inputfile, personality=personality)
        convertor = prop2po.prop2po(personality=personality)
        if proptemplate:
            templatefile = BytesIO(proptemplate.encode())
            templateprop = properties.propfile(templatefile)
            outputpo = convertor.mergestore(templateprop, inputprop)
        else:
            outputpo = convertor.convertstore(inputprop)
        return outputpo

    @staticmethod
    def convertprop(propsource):
        """Call the convertprop, return the outputfile."""
        inputfile = BytesIO(propsource.encode())
        outputfile = BytesIO()
        templatefile = None
        assert prop2po.convertprop(inputfile, outputfile, templatefile)
        return outputfile.getvalue()

    @staticmethod
    def singleelement(pofile):
        """Checks that the pofile contains a single non-header element, and returns it."""
        assert len(pofile.units) == 2
        assert pofile.units[0].isheader()
        print(pofile)
        return pofile.units[1]

    @staticmethod
    def countelements(pofile):
        """Counts the number of non-header entries."""
        assert pofile.units[0].isheader()
        print(pofile)
        return len(pofile.units) - 1

    def test_simpleentry(self) -> None:
        """Checks that a simple properties entry converts properly to a po entry."""
        propsource = "SAVEENTRY=Save file\n"
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.source == "Save file"
        assert pounit.target == ""

    def test_convertprop(self) -> None:
        """Checks that the convertprop function is working."""
        propsource = "SAVEENTRY=Save file\n"
        posource = self.convertprop(propsource)
        pofile = po.pofile(BytesIO(posource))
        pounit = self.singleelement(pofile)
        assert pounit.source == "Save file"
        assert pounit.target == ""

    def test_no_value_entry(self) -> None:
        """Checks that a properties entry without value is converted."""
        propsource = "KEY = \n"
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.getcontext() == "KEY"
        assert pounit.source == ""
        assert pounit.target == ""

    def test_no_separator_entry(self) -> None:
        """Checks that a properties entry without separator is converted."""
        propsource = "KEY\n"
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.getcontext() == "KEY"
        assert pounit.source == ""
        assert pounit.target == ""

    def test_value_no_key_entry(self) -> None:
        """Checks that a properties entry with value but no key is converted."""
        propsource = "=value\n"
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        # Empty key should use special <empty> marker for location
        assert pounit.getlocations() == ["<empty>"]
        assert pounit.source == "value"
        assert pounit.target == ""

    def test_tab_at_end_of_string(self) -> None:
        """Check that we preserve tabs at the end of a string."""
        propsource = r"TAB_AT_END=This sentence has a tab at the end.\t"
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.source == "This sentence has a tab at the end.\t"

        propsource = (
            r"SPACE_THEN_TAB_AT_END=This sentence has a space then tab at the end. \t"
        )
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.source == "This sentence has a space then tab at the end. \t"

        propsource = (
            r"SPACE_AT_END=This sentence will keep its 4 spaces at the end.    "
        )
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.source == "This sentence will keep its 4 spaces at the end.    "

        propsource = r"SPACE_AT_END_NO_TRIM=This sentence will keep its 4 spaces at the end.\    "
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.source == "This sentence will keep its 4 spaces at the end.    "

        propsource = r"SPACE_AT_END_NO_TRIM2=This sentence will keep its 4 spaces at the end.\\    "
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.source == "This sentence will keep its 4 spaces at the end.\\    "

    def test_tab_at_start_of_value(self) -> None:
        """Check that tabs in a property are ignored where appropriate."""
        propsource = r"property	=	value"
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.getlocations()[0] == "property"
        assert pounit.source == "value"

    def test_unicode(self) -> None:
        """Checks that unicode entries convert properly."""
        unistring = r"Norsk bokm\u00E5l"
        propsource = f"nb = {unistring}\n"
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        print(repr(pofile.units[0].target))
        print(repr(pounit.source))
        assert pounit.source == "Norsk bokm\u00e5l"

    def test_multiline_escaping(self) -> None:
        """Checks that multiline entries can be parsed."""
        propsource = r"""5093=Unable to connect to your IMAP server. You may have exceeded the maximum number \
of connections to this server. If so, use the Advanced IMAP Server Settings dialog to \
reduce the number of cached connections."""
        pofile = self.prop2po(propsource)
        print(repr(pofile.units[1].target))
        assert self.countelements(pofile) == 1

    def test_comments(self) -> None:
        """Test to ensure that we take comments from .properties and place them in .po."""
        propsource = """# Comment
prefPanel-smime=Security"""
        pofile = self.prop2po(propsource)
        pounit = self.singleelement(pofile)
        assert pounit.getnotes("developer") == "Comment"

    def test_multiline_comments(self) -> None:
        """Test to ensure that we handle multiline comments well."""
        propsource = """# Comment
# commentary 2

## @name GENERIC_ERROR
## @loc none
prefPanel-smime=
"""
        pofile = self.prop2po(propsource)
        print(bytes(pofile))
        # header comments:
        assert b"#. # Comment\n#. # commentary 2" in bytes(pofile)
        pounit = self.singleelement(pofile)
        assert pounit.getnotes("developer") == "# @name GENERIC_ERROR\n# @loc none"

    def test_folding_accesskeys(self) -> None:
        """Check that we can fold various accesskeys into their associated label (bug #115)."""
        propsource = r"""cmd_addEngine.label = Add Engines...
cmd_addEngine.accesskey = A"""
        pofile = self.prop2po(propsource, personality="mozilla")
        pounit = self.singleelement(pofile)
        assert pounit.source == "&Add Engines..."

    def test_dont_translate(self) -> None:
        """Check that we know how to ignore don't translate instructions in properties files (bug #116)."""
        propsource = """# LOCALIZATION NOTE (dont): DONT_TRANSLATE.
dont=don't translate me
do=translate me
"""
        pofile = self.prop2po(propsource)
        assert self.countelements(pofile) == 1

    def test_duplicate_locations_mozilla(self) -> None:
        """
        Check that we handle duplicate locations properly in Mozilla properties files.
        This tests the fix for the Lithuanian Firefox recovery issue where duplicate
        locations could cause a traceback.

        When there are duplicate keys, locationindex will contain only the first
        occurrence from the translated file. Both template entries are processed,
        but both will map to the same translation (the first occurrence).
        """
        # Template file with duplicate keys
        templateprop = """key1=Original value 1
key1=Original value 2
key2=Another value
"""
        # Translation file with duplicate keys
        translatedprop = """key1=Translated value 1
key1=Translated value 2
key2=Translated value
"""
        # Convert using Mozilla personality - should not raise AttributeError
        pofile = self.prop2po(translatedprop, templateprop, personality="mozilla")

        # We get 3 units because both key1 entries from template are processed
        # Both map to the first translated value due to locationindex behavior
        assert self.countelements(pofile) == 3

        # First key1 entry
        pounit1 = pofile.units[1]
        assert pounit1.source == "Original value 1"
        assert pounit1.target == "Translated value 1"
        assert pounit1.getlocations() == ["key1"]

        # Second key1 entry (duplicate location)
        pounit2 = pofile.units[2]
        assert pounit2.source == "Original value 2"
        # Both key1 entries map to first translated value
        assert pounit2.target == "Translated value 1"
        assert pounit2.getlocations() == ["key1"]

        # key2 entry
        pounit3 = pofile.units[3]
        assert pounit3.source == "Another value"
        assert pounit3.target == "Translated value"
        assert pounit3.getlocations() == ["key2"]

    def test_emptyproperty(self) -> None:
        """Checks that empty property definitions survive into po file, bug 15."""
        for delimiter in ["=", ""]:
            propsource = f"# comment\ncredit{delimiter}"
            pofile = self.prop2po(propsource)
            pounit = self.singleelement(pofile)
            assert pounit.getlocations() == ["credit"]
            assert pounit.getcontext() == "credit"
            assert 'msgctxt "credit"' in str(pounit)
            assert b"#. comment" in bytes(pofile)
            assert pounit.source == ""

    def test_emptyproperty_translated(self) -> None:
        """Checks that if we translate an empty property it makes it into the PO."""
        for delimiter in ["=", ""]:
            proptemplate = f"credit{delimiter}"
            propsource = "credit=Translators Names"
            pofile = self.prop2po(propsource, proptemplate)
            pounit = self.singleelement(pofile)
            assert pounit.getlocations() == ["credit"]
            # FIXME we don't seem to get a _: comment but we should
            # assert pounit.getcontext() == "credit"
            assert pounit.source == ""
            assert pounit.target == "Translators Names"

    def test_newlines_in_value(self) -> None:
        """Check that we can carry newlines that appear in the property value into the PO."""
        propsource = """prop=\\nvalue\\n\n"""
        pofile = self.prop2po(propsource)
        unit = self.singleelement(pofile)
        assert unit.source == "\nvalue\n"

    def test_header_comments(self) -> None:
        """Check that we can handle comments not directly associated with a property."""
        propsource = """# Header comment\n\n# Comment\n\nprop=value\n"""
        pofile = self.prop2po(propsource)
        unit = self.singleelement(pofile)
        assert unit.source == "value"
        assert unit.getnotes("developer") == "Comment"

    def test_unassociated_comment_order(self) -> None:
        """Check that we can handle the order of unassociated comments."""
        propsource = """# Header comment\n\n# 1st Unassociated comment\n\n# 2nd Connected comment\nprop=value\n"""
        pofile = self.prop2po(propsource)
        unit = self.singleelement(pofile)
        assert unit.source == "value"
        assert (
            unit.getnotes("developer")
            == "1st Unassociated comment\n\n2nd Connected comment"
        )

    def test_x_header(self) -> None:
        """
        Test that we correctly create the custom header entries
        (accelerators, merge criterion).
        """
        propsource = """prop=value\n"""

        outputpo = self.prop2po(propsource, personality="mozilla")
        assert b"X-Accelerator-Marker" in bytes(outputpo)
        assert b"X-Merge-On" in bytes(outputpo)

        # Even though the gaia flavour inherits from mozilla, it should not
        # get the header
        outputpo = self.prop2po(propsource, personality="gaia")
        assert b"X-Accelerator-Marker" not in bytes(outputpo)
        assert b"X-Merge-On" not in bytes(outputpo)

    def test_gaia_plurals(self) -> None:
        """Test conversion of gaia plural units."""
        propsource = """
message-multiedit-header={[ plural(n) ]}
message-multiedit-header[zero]=Edit
message-multiedit-header[one]={{ n }} selected
message-multiedit-header[two]={{ n }} selected
message-multiedit-header[few]={{ n }} selected
message-multiedit-header[many]={{ n }} selected
message-multiedit-header[other]={{ n }} selected
"""
        outputpo = self.prop2po(propsource, personality="gaia")
        pounit = outputpo.units[-1]
        assert pounit.hasplural()
        assert pounit.getlocations() == ["message-multiedit-header"]

        print(outputpo)
        zero_unit = outputpo.units[-2]
        assert not zero_unit.hasplural()
        assert zero_unit.source == "Edit"

    def test_successive_gaia_plurals(self) -> None:
        """Test conversion of two successive gaia plural units."""
        propsource = """
message-multiedit-header={[ plural(n) ]}
message-multiedit-header[zero]=Edit
message-multiedit-header[one]={{ n }} selected
message-multiedit-header[two]={{ n }} selected
message-multiedit-header[few]={{ n }} selected
message-multiedit-header[many]={{ n }} selected
message-multiedit-header[other]={{ n }} selected

message-multiedit-header2={[ plural(n) ]}
message-multiedit-header2[zero]=Edit 2
message-multiedit-header2[one]={{ n }} selected 2
message-multiedit-header2[two]={{ n }} selected 2
message-multiedit-header2[few]={{ n }} selected 2
message-multiedit-header2[many]={{ n }} selected 2
message-multiedit-header2[other]={{ n }} selected 2
"""
        outputpo = self.prop2po(propsource, personality="gaia")
        pounit = outputpo.units[-1]
        assert pounit.hasplural()
        assert pounit.getlocations() == ["message-multiedit-header2"]

        pounit = outputpo.units[-3]
        assert pounit.hasplural()
        assert pounit.getlocations() == ["message-multiedit-header"]

        print(outputpo)
        zero_unit = outputpo.units[-2]
        assert not zero_unit.hasplural()
        assert zero_unit.source == "Edit 2"

        zero_unit = outputpo.units[-4]
        assert not zero_unit.hasplural()
        assert zero_unit.source == "Edit"

    def test_duplicate_keys(self) -> None:
        """Check that we correctly handle duplicate keys."""
        source = """
key=value
key=value
"""
        po_file = self.prop2po(source)
        assert self.countelements(po_file) == 1
        po_unit = self.singleelement(po_file)
        assert po_unit.source == "value"

        source = """
key=value
key=another value
"""
        po_file = self.prop2po(source)
        assert self.countelements(po_file) == 2
        po_unit = po_file.units[1]
        assert po_unit.source == "value"
        assert po_unit.getlocations() == ["key"]
        po_unit = po_file.units[2]
        assert po_unit.source == "another value"
        assert po_unit.getlocations() == ["key"]

        source = """
key1=value
key2=value
"""
        po_file = self.prop2po(source)
        assert self.countelements(po_file) == 2
        po_unit = po_file.units[1]
        assert po_unit.source == "value"
        assert po_unit.getlocations() == ["key1"]
        po_unit = po_file.units[2]
        assert po_unit.source == "value"
        assert po_unit.getlocations() == ["key2"]

    def test_gwt_plurals(self) -> None:
        """Test conversion of gwt plural units."""
        propsource = """
message-multiedit-header={0,number} selected
message-multiedit-header[none]=Edit
message-multiedit-header[one]={0,number} selected
message-multiedit-header[two]={0,number} selected
message-multiedit-header[few]={0,number} selected
message-multiedit-header[many]={0,number} selected
"""
        outputpo = self.prop2po(propsource, personality="gwt")
        pounit = outputpo.units[-1]
        assert pounit.getlocations() == ["message-multiedit-header"]

    def test_strings_bilingual_simple(self) -> None:
        """Test that .strings files are treated as bilingual in convertstore mode."""
        propsource = r""""Source text" = "Translated text";"""
        # .strings files use UTF-16 encoding
        inputfile = BytesIO(propsource.encode("utf-16"))
        inputprop = properties.propfile(inputfile, personality="strings")
        convertor = prop2po.prop2po(personality="strings")
        outputpo = convertor.convertstore(inputprop)
        pounit = self.singleelement(outputpo)
        # After fix, source should be left side (key) and target should be right side (value)
        assert pounit.source == "Source text", (
            f"Expected 'Source text' but got {pounit.source!r}"
        )
        assert pounit.target == "Translated text", (
            f"Expected 'Translated text' but got {pounit.target!r}"
        )

    def test_strings_bilingual_multiline(self) -> None:
        """Test multiline .strings bilingual conversion (issue from bug report)."""
        # This is based on the actual bug report
        propsource = r"""/* Overwrite of the app folder */
"A '<AppNameRemovedForPrivacy>' folder has been found in  your Music folder. Unfortunately it seems incomplete.\n\nTo restart <AppNameRemovedForPrivacy> synchronization, trash the ~/Music/<AppNameRemovedForPrivacy> folder and relaunch the application." = "An incomplete <AppNameRemovedForPrivacy> folder has been found in your Music folder.\n\nTo restart, delete the ~/Music/<AppNameRemovedForPrivacy> folder and relaunch the application.";
"""
        # .strings files use UTF-16 encoding
        inputfile = BytesIO(propsource.encode("utf-16"))
        inputprop = properties.propfile(inputfile, personality="strings")
        convertor = prop2po.prop2po(personality="strings")
        outputpo = convertor.convertstore(inputprop)
        pounit = self.singleelement(outputpo)
        # The msgid should be the source (left side), not the translation (right side)
        assert "A '<AppNameRemovedForPrivacy>' folder has been found" in pounit.source
        assert "Unfortunately it seems incomplete" in pounit.source
        # The msgstr should be the translation (right side)
        assert "An incomplete <AppNameRemovedForPrivacy> folder" in pounit.target
        assert "To restart, delete" in pounit.target
        # The location should be the key (not URL-encoded full text)
        locations = pounit.getlocations()
        assert len(locations) == 1
        # Location should be a reasonable identifier, not the full multiline text
        assert "A '<AppNameRemovedForPrivacy>' folder" in locations[0]

    def test_strings_merge_mode_uses_template_value_as_msgid(self) -> None:
        """Ensure merge mode keeps standard monolingual behavior for .strings."""
        templatesource = r""""Hello" = "Hello from template";"""
        translationsource = r""""Hello" = "Hola des de la plantilla";"""

        templatefile = BytesIO(templatesource.encode("utf-16"))
        translationfile = BytesIO(translationsource.encode("utf-16"))
        templateprop = properties.propfile(templatefile, personality="strings")
        translationprop = properties.propfile(translationfile, personality="strings")

        convertor = prop2po.prop2po(personality="strings")
        outputpo = convertor.mergestore(templateprop, translationprop)
        pounit = self.singleelement(outputpo)

        # In merge mode, msgid comes from the template value and msgstr from input value.
        assert pounit.source == "Hello from template"
        assert pounit.target == "Hola des de la plantilla"
        assert pounit.getlocations() == ["Hello"]

    def test_strings_merge_template(self) -> None:
        """Test that .strings merge mode uses template value as msgid."""
        template_source = b'"greeting" = "Hello";\n"farewell" = "Goodbye";\n'
        input_source = b'"greeting" = "Hola";\n"farewell" = "Adeu";\n'

        template_file = BytesIO(template_source)
        input_file = BytesIO(input_source)

        template_prop = properties.propfile(template_file, personality="strings-utf8")
        input_prop = properties.propfile(input_file, personality="strings-utf8")

        convertor = prop2po.prop2po(personality="strings-utf8")
        outputpo = convertor.mergestore(template_prop, input_prop)

        units = [u for u in outputpo.units if u.source]
        assert len(units) == 2
        # msgid must be the English value from the template, not the key
        assert units[0].source == "Hello"
        assert units[0].target == "Hola"
        assert units[1].source == "Goodbye"
        assert units[1].target == "Adeu"


class TestProp2POCommand(test_convert.TestConvertCommand, TestProp2PO):
    """Tests running actual prop2po commands on files."""

    convertmodule = prop2po
    defaultoptions = {"progress": "none"}
    expected_options = [
        "-P, --pot",
        "-t TEMPLATE, --template=TEMPLATE",
        "--personality=TYPE",
        "--encoding=ENCODING",
        "--duplicates=DUPLICATESTYLE",
    ]



from translate.lang.team import guess_language


def test_simple() -> None:
    """
    Test the regex, team snippet and language name snippets at a high
    level.
    """
    # standard regex guess
    assert guess_language("ab@li.org") == "ab"
    # We never suggest 'en', it's always a mistake
    assert guess_language("en@li.org") is None
    # We can't have a single char language code
    assert guess_language("C@li.org") is None
    # Testing regex postfilter
    assert guess_language("LL@li.org") is None

    # snippet guess based on contact info
    assert guess_language("assam@mm.assam-glug.org") == "as"
    # snippet guess based on a language name
    assert guess_language("Hawaiian") == "haw"

    # We found nothing
    assert guess_language("Bork bork") is None
